"""
Manipulate atoms and molecules with mouse interactively
based on force field methods.


Molecule: Rigid molecules, to constrain all internal
degrees of freedom using the RATTLE-type constraints of
the The FixBondLengths class to constrain all internal atomic distances.

Intermolecular: atomic radii.


"""

import bpy
import numpy as np
from batoms.utils.butils import (
    get_selected_batoms,
    get_selected_objects,
    get_selected_vertices,
    object_mode,
)
from batoms import Batoms
from .modal_rigid_body import mouse2positions
from bpy.props import (
    StringProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
)

from ase.optimize import QuasiNewton

try:
    from asap3 import EMT
except ImportError:
    from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
import logging

logger = logging.getLogger(__name__)


def translate(selected_vertices, displacement, fmax=0.05, steps=5, frame_start=0):
    """ """
    if len(selected_vertices) == 0:
        return
    displacement = displacement[:3] * 0.03
    label = selected_vertices[0][0]
    object_mode()
    # bpy.context.scene.frame_set(frame_start)
    batoms = Batoms(label)
    # Move selected atom with mouse
    atoms = batoms.atoms
    fixed = []
    for label, species, name, index in selected_vertices:
        # batom = Batom(name)
        ind = np.where(atoms.arrays["species"] == species)[0]
        # positions = batom.positions
        atoms.positions[ind[index]] += displacement
        fixed.append(FixAtoms(indices=ind[index]))
    # run optimize
    # print(atoms.positions)
    atoms.constraints = fixed
    optimize(atoms, fmax, steps)
    batoms.set_frames([atoms], frame_start=frame_start)
    # set new positions of atoms
    batoms.positions = atoms
    # batoms.model_style = 1
    # batoms.bondsetting.add(['Al', 'Al'])
    # batoms.draw_bonds()
    # batoms.draw()


def optimize(atoms, fmax=0.05, steps=5):
    """ """
    atoms.calc = EMT()
    qn = QuasiNewton(atoms)
    qn.run(fmax=fmax, steps=steps)


def add_constraint(atoms, mol_index):
    """ """
    from ase.constraints import FixAtoms, FixBondLengths

    # RATTLE-type constraints on Molecule.
    atoms.constraints = FixBondLengths(
        [(3 * i + j, 3 * i + (j + 1) % 3) for i in range(3**3) for j in [0, 1, 2]]
    )
    #
    # Fix others
    mask = set(range(len(atoms))) - mol_index
    fixlayers = FixAtoms(mask=mask)
    atoms.set_constraint(fixlayers)
    return atoms


class Force_Field_Operator(bpy.types.Operator):
    """Force field"""

    bl_idname = "object.force_field_operator"
    bl_label = "Force field translate Operator"

    mouse_position: FloatVectorProperty(size=4, default=(0, 0, 0, 0))
    nframe: IntProperty(default=0)
    previous: StringProperty()

    @property
    def selected_batoms(self):
        return get_selected_batoms()

    @property
    def selected_batom(self):
        return get_selected_objects("batom")

    @property
    def selected_vertices(self):
        return get_selected_vertices()

    def modal(self, context, event):
        """ """
        if event.type in {"ESC"}:
            return {"CANCELLED"}
        elif event.type == "MIDDLEMOUSE":
            self.previous = "MIDDLEMOUSE"
            return {"PASS_THROUGH"}
        elif event.type == "MOUSEMOVE":
            mouse_position = np.array([event.mouse_x, event.mouse_y, 0, 0])
            if self.previous == "MIDDLEMOUSE":
                self.mouse_position = mouse_position
                self.previous = "MOUSEMOVE"
                return {"RUNNING_MODAL"}
            selected_vertices = self.selected_vertices
            if len(selected_vertices) != 1:
                self.mouse_position = mouse_position
                return {"RUNNING_MODAL"}
            delta = mouse_position - self.mouse_position
            displacement = mouse2positions(
                delta, self.viewports_3D.spaces.active.region_3d.view_matrix
            )
            self.mouse_position = mouse_position
            ffpanel = context.scene.ffpanel
            fmax = ffpanel.fmax
            steps = ffpanel.steps
            translate(
                selected_vertices, displacement, fmax, steps, frame_start=self.nframe
            )
            self.nframe += 1
            self.previous = "MOUSEMOVE"
            return {"RUNNING_MODAL"}
        else:
            return {"PASS_THROUGH"}

    def invoke(self, context, event):
        """ """
        if context.object:
            for area in bpy.context.screen.areas:
                if area.type == "VIEW_3D":
                    self.viewports_3D = area
            self.mouse_position = np.array([event.mouse_x, event.mouse_y, 0, 0])
            self.nframe = 0
            context.window_manager.modal_handler_add(self)
            return {"RUNNING_MODAL"}
        else:
            self.report({"WARNING"}, "No active object, could not finish")
            return {"CANCELLED"}


class Force_Field_Modal_Panel(bpy.types.Panel):
    bl_idname = "force_field_modal_panel"
    bl_label = "Force field"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Real"
    bl_idname = "FORCEFIELD_PT_Modal"

    def draw(self, context):
        ffpanel = context.scene.ffpanel
        layout = self.layout
        box = layout.box()
        row = box.row()
        row.prop(ffpanel, "fmax", expand=True)
        row = box.row()
        row.prop(ffpanel, "steps", expand=True)
        layout = self.layout
        layout.operator(
            "object.force_field_operator", text="Force field", icon="FACESEL"
        )


class ForceFieldProperties(bpy.types.PropertyGroup):
    @property
    def selected_batoms(self):
        return get_selected_batoms()

    def Callback_modify_fmax(self, context):
        # clpanel = bpy.context.scene.clpanel
        # transform = clpanel.transform
        # modify_transform(self.selected_batoms, transform)
        pass

    def Callback_modify_steps(self, context):
        # clpanel = bpy.context.scene.clpanel
        # transform = clpanel.transform
        # modify_transform(self.selected_batoms, transform)
        pass

    fmax: FloatProperty(
        name="Max force",
        default=0.05,
        description="max force",
        update=Callback_modify_fmax,
    )
    steps: IntProperty(
        name="Max steps",
        default=5,
        description="Max steps",
        update=Callback_modify_steps,
    )


def register():
    bpy.utils.register_class(Force_Field_Operator)
    bpy.utils.register_class(Force_Field_Modal_Panel)


def unregister():
    bpy.utils.unregister_class(Force_Field_Operator)
    bpy.utils.unregister_class(Force_Field_Modal_Panel)


if __name__ == "__main__":
    register()

    # test call
    bpy.ops.object.modal_operator("INVOKE_DEFAULT")



import bpy
from bpy_extras import view3d_utils
import blf
import gpu
from gpu_extras.batch import batch_for_shader
import numpy as np


class DrawText:
    def __init__(self, context, label):
        self.label = label

    def draw_callback_text(self, positions, texts):
        # get the context arguments
        context = bpy.context
        region = context.region
        rv3d = context.region_data
        font_id = 0
        n = len(positions)
        for i in range(n):
            coord = positions[i]
            coord_2d = view3d_utils.location_3d_to_region_2d(
                region, rv3d, coord, default=None
            )
            # draw some text
            blf.position(font_id, coord_2d[0], coord_2d[1], 0)
            blf.size(font_id, 20, 72)
            blf.draw(font_id, str(texts[i]))

    def add_handle(self, positions, texts):
        self._handle_label = bpy.types.SpaceView3D.draw_handler_add(
            self.draw_callback_text, (positions, texts), "WINDOW", "POST_PIXEL"
        )

    def remove_handle(self):
        if hasattr(self, "_handle_label"):
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_label, "WINDOW")
            delattr(self, "_handle_label")


class DrawCrystalAxes:
    def __init__(self, context, label):
        self.label = label
        self.length = 50
        self.texts = ["a", "b", "c"]

    def draw_callback_text(self, cell):
        # get the context arguments
        context = bpy.context
        region = context.region
        rv3d = context.region_data
        font_id = 0
        origin = view3d_utils.location_3d_to_region_2d(
            region, rv3d, [0, 0, 0], default=None
        )
        for i in range(3):
            color = [0, 0, 0, 1]
            color[i] = 1
            coord = cell[i]
            coord_2d = np.array(
                view3d_utils.location_3d_to_region_2d(region, rv3d, coord, default=None)
                - origin  # noqa W503
            )
            vector = coord_2d / np.linalg.norm(coord_2d) * self.length + [100, 100]
            positions = [[100, 100], vector]
            # draw some text
            blf.position(font_id, vector[0], vector[1], 0)
            blf.size(font_id, 20, 72)
            blf.draw(font_id, str(self.texts[i]))
            # 50% alpha, 2 pixel width line
            shader = gpu.shader.from_builtin("2D_UNIFORM_COLOR")
            gpu.state.blend_set("ALPHA")
            gpu.state.line_width_set(4.0)
            batch = batch_for_shader(shader, "LINE_STRIP", {"pos": positions})
            shader.bind()
            shader.uniform_float("color", tuple(color))
            batch.draw(shader)
            # restore opengl defaults
            gpu.state.line_width_set(1.0)
            gpu.state.blend_set("NONE")

    def add_handle(self, cell):
        self._handle_label = bpy.types.SpaceView3D.draw_handler_add(
            self.draw_callback_text, (cell,), "WINDOW", "POST_PIXEL"
        )

    def remove_handle(self):
        if hasattr(self, "_handle_label"):
            bpy.types.SpaceView3D.draw_handler_remove(self._handle_label, "WINDOW")
            delattr(self, "_handle_label")



import bpy
import numpy as np
from batoms.base.object import ObjectGN
from batoms.utils.butils import object_mode, compareNodeType, get_node_by_name
from batoms.utils import number2String, string2Number
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)

default_attributes = [
    {"name": "atoms_index", "data_type": "INT"},
    {"name": "species_index", "data_type": "INT"},
    {"name": "show", "data_type": "INT"},
    {"name": "select", "data_type": "INT"},
    {"name": "model_style", "data_type": "INT"},
    {"name": "scale", "data_type": "FLOAT"},
]

default_GroupInput = [
    ["atoms_index", "NodeSocketInt"],
    ["species_index", "NodeSocketInt"],
    ["show", "NodeSocketBool"],
    ["select", "NodeSocketInt"],
    ["model_style", "NodeSocketInt"],
    ["scale", "NodeSocketFloat"],
]


default_search_bond_datas = {
    "atoms_index": np.ones(0, dtype=int),
    "species_index": np.ones(0, dtype=int),
    "species": np.ones(0, dtype="U4"),
    "positions": np.zeros((0, 3)),
    "scales": np.zeros(0),
    "offsets": np.zeros((0, 3)),
    "model_styles": np.ones(0, dtype=int),
    "shows": np.ones(0, dtype=int),
    "selects": np.ones(0, dtype=int),
}


class SearchBond(ObjectGN):
    def __init__(
        self,
        label=None,
        search_bond_datas=None,
        batoms=None,
        load=False,
    ):
        """SearchBond Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            search_bond_datas (_type_, optional): _description_. Defaults to None.
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        name = "search_bond"
        ObjectGN.__init__(self, label, name)
        if not load or not self.loadable():
            if search_bond_datas is not None:
                self.build_object(search_bond_datas)
            else:
                self.build_object(default_search_bond_datas)

    def loadable(self):
        """Check loadable or not"""
        # object exist
        obj = bpy.data.objects.get(self.obj_name)
        if obj is None:
            return False
        # batoms exist, and flag is True
        # coll = bpy.data.collections.get(self.label)
        # if coll is None:
        # return False
        # return coll.Bbond.flag
        return True

    def build_object(self, datas, attributes={}):
        """
        build child object and add it to main objects.
        """
        # tstart = time()
        if len(datas["positions"].shape) == 2:
            self.trajectory = {
                "positions": np.array([datas["positions"]]),
                "offsets": np.array([datas["offsets"]]),
            }
            positions = datas["positions"]
            offsets = datas["offsets"]
        elif len(datas["positions"].shape) == 3:
            self.trajectory = {
                "positions": datas["positions"],
                "offsets": datas["offsets"],
            }
            positions = datas["positions"][0]
            offsets = datas["offsets"][0]
        else:
            raise Exception("Shape of positions is wrong!")
        #
        attributes.update(
            {
                "atoms_index": datas["atoms_index"],
                "species_index": datas["species_index"],
                "show": datas["shows"],
                "model_style": datas["model_styles"],
                "select": datas["selects"],
                "scale": datas["scales"],
            }
        )
        name = "%s_search_bond" % self.label
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(positions, [], [])
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        # Add attributes
        for att in default_attributes:
            self.add_attribute(**att)
        self.batoms.coll.objects.link(obj)
        obj.batoms.type = "BOND"
        obj.batoms.label = self.batoms.label
        obj.Bbond.label = self.batoms.label
        obj.parent = self.batoms.obj
        #
        name = "%s_search_bond_offset" % self.label
        self.delete_obj(name)
        mesh = bpy.data.meshes.new(name)
        mesh.from_pydata(offsets, [], [])
        mesh.update()
        obj = bpy.data.objects.new(name, mesh)
        self.batoms.coll.objects.link(obj)
        obj.hide_set(True)
        obj.parent = self.obj
        bpy.context.view_layer.update()
        self.set_attributes(attributes)
        self.init_geometry_node_modifier(default_GroupInput)
        self.build_geometry_node()
        self.set_trajectory(self.trajectory)
        # print('boundary: build_object: {0:10.2f} s'.format(time() - tstart))

    def update(self, bondlist, mollists, moldatas, arrays, cell):
        self.hide = False
        search_bond_data = self.calc_search_bond_data(
            bondlist, mollists, moldatas, arrays, cell
        )
        self.set_arrays(search_bond_data)

    def build_geometry_node(self):
        """ """
        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        GroupOutput = nodes[1]
        # ------------------------------------------------------------------
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        links.new(GroupInput.outputs["Geometry"], JoinGeometry.inputs["Geometry"])
        links.new(JoinGeometry.outputs["Geometry"], GroupOutput.inputs["Geometry"])
        # ------------------------------------------------------------------
        # transform postions of batoms to boundary
        ObjectBatoms = get_node_by_name(
            nodes, "%s_ObjectBatoms" % self.label, "GeometryNodeObjectInfo"
        )
        ObjectBatoms.inputs["Object"].default_value = self.batoms.obj
        PositionBatoms = get_node_by_name(
            nodes, "%s_PositionBatoms" % (self.label), "GeometryNodeInputPosition"
        )
        TransferBatoms = get_node_by_name(
            nodes, "%s_TransferBatoms" % (self.label), "GeometryNodeSampleIndex"
        )
        TransferBatoms.data_type = "FLOAT_VECTOR"
        links.new(ObjectBatoms.outputs["Geometry"], TransferBatoms.inputs[0])
        links.new(PositionBatoms.outputs["Position"], TransferBatoms.inputs[1])
        links.new(GroupInput.outputs[1], TransferBatoms.inputs["Index"])
        # ------------------------------------------------------------------
        # add positions with offsets
        # transfer offsets from object self.obj_o
        ObjectOffsets = get_node_by_name(
            nodes, "%s_ObjectOffsets" % (self.label), "GeometryNodeObjectInfo"
        )
        ObjectOffsets.inputs["Object"].default_value = self.obj_o
        PositionOffsets = get_node_by_name(
            nodes, "%s_PositionOffsets" % (self.label), "GeometryNodeInputPosition"
        )
        TransferOffsets = get_node_by_name(
            nodes, "%s_TransferOffsets" % self.label, "GeometryNodeSampleIndex"
        )
        TransferOffsets.data_type = "FLOAT_VECTOR"
        links.new(ObjectOffsets.outputs["Geometry"], TransferOffsets.inputs[0])
        links.new(PositionOffsets.outputs["Position"], TransferOffsets.inputs[1])
        OffsetNode = self.vectorDotMatrix(
            self.gn_node_group, TransferOffsets.outputs[2], self.batoms.cell, ""
        )
        # we need one add operation to get the positions with offset
        VectorAdd = get_node_by_name(
            nodes, "%s_VectorAdd" % (self.label), "ShaderNodeVectorMath"
        )
        VectorAdd.operation = "ADD"
        links.new(TransferBatoms.outputs[2], VectorAdd.inputs[0])
        links.new(OffsetNode.outputs[0], VectorAdd.inputs[1])
        # set positions
        SetPosition = get_node_by_name(
            nodes, "%s_SetPosition" % self.label, "GeometryNodeSetPosition"
        )
        links.new(GroupInput.outputs["Geometry"], SetPosition.inputs["Geometry"])
        links.new(VectorAdd.outputs[0], SetPosition.inputs["Position"])
        #
        # ------------------------------------------------------------------
        # transform scale of batoms to boundary
        if bpy.app.version_string >= "3.2.0":
            ScaleBatoms = get_node_by_name(
                nodes,
                "%s_ScaleBatoms" % (self.label),
                "GeometryNodeInputNamedAttribute",
            )
            # need to be "FLOAT_VECTOR", because scale is "FLOAT_VECTOR"
            ScaleBatoms.data_type = "FLOAT_VECTOR"
            ScaleBatoms.inputs[0].default_value = "scale"
            TransferScale = get_node_by_name(
                nodes, "%s_TransferScale" % (self.label), "GeometryNodeSampleIndex"
            )
            TransferScale.data_type = "FLOAT_VECTOR"
            links.new(ObjectBatoms.outputs["Geometry"], TransferScale.inputs[0])
            links.new(ScaleBatoms.outputs["Attribute"], TransferScale.inputs[1])
            links.new(GroupInput.outputs[1], TransferScale.inputs["Index"])

    def add_geometry_node(self, spname):
        """ """
        nodes = self.gn_node_group.nodes
        links = self.gn_node_group.links
        GroupInput = nodes[0]
        SetPosition = get_node_by_name(nodes, "%s_SetPosition" % self.label)
        JoinGeometry = get_node_by_name(
            nodes, "%s_JoinGeometry" % self.label, "GeometryNodeJoinGeometry"
        )
        CompareSpecies = get_node_by_name(
            nodes, "CompareFloats_%s_%s" % (self.label, spname), compareNodeType
        )
        CompareSpecies.operation = "EQUAL"
        # CompareSpecies.data_type = 'INT'
        CompareSpecies.inputs[1].default_value = string2Number(spname)
        InstanceOnPoint = get_node_by_name(
            nodes,
            "InstanceOnPoint_%s_%s" % (self.label, spname),
            "GeometryNodeInstanceOnPoints",
        )
        ObjectInfo = get_node_by_name(
            nodes, "ObjectInfo_%s_%s" % (self.label, spname), "GeometryNodeObjectInfo"
        )
        ObjectInfo.inputs["Object"].default_value = self.batoms.species.instancers[
            spname
        ]
        BoolShow = get_node_by_name(
            nodes,
            "BooleanMath_%s_%s_1" % (self.label, spname),
            "FunctionNodeBooleanMath",
        )
        #
        links.new(SetPosition.outputs["Geometry"], InstanceOnPoint.inputs["Points"])
        links.new(GroupInput.outputs[2], CompareSpecies.inputs[0])
        links.new(GroupInput.outputs[3], BoolShow.inputs[0])
        # transfer scale
        if bpy.app.version_string >= "3.2.0":
            TransferScale = get_node_by_name(
                nodes, "%s_TransferScale" % (self.label), "GeometryNodeSampleIndex"
            )
            links.new(TransferScale.outputs[2], InstanceOnPoint.inputs["Scale"])
        else:
            links.new(GroupInput.outputs[6], InstanceOnPoint.inputs["Scale"])
        links.new(CompareSpecies.outputs[0], BoolShow.inputs[1])
        links.new(BoolShow.outputs["Boolean"], InstanceOnPoint.inputs["Selection"])
        links.new(ObjectInfo.outputs["Geometry"], InstanceOnPoint.inputs["Instance"])
        links.new(InstanceOnPoint.outputs["Instances"], JoinGeometry.inputs["Geometry"])

    def update_geometry_node_instancer(self, spname, instancer):
        """When instances are re-build, we need also update
        the geometry node.

        Args:
            spname (str): name of the species
        """
        from batoms.utils.butils import get_node_by_name

        # update  instancers
        ObjectInfo = get_node_by_name(
            self.gn_node_group.nodes,
            "ObjectInfo_%s_%s" % (self.label, spname),
            "GeometryNodeObjectInfo",
        )
        ObjectInfo.inputs["Object"].default_value = instancer
        logger.debug("update boundary instancer: {}".format(spname))

    @property
    def obj_o(self):
        return self.get_obj_o()

    def get_obj_o(self):
        name = "%s_search_bond_offset" % self.label
        obj_o = bpy.data.objects.get(name)
        if obj_o is None:
            raise KeyError("%s object is not exist." % name)
        return obj_o

    def get_search_bond(self):
        boundary = np.array(self.batoms.obj.batoms.boundary)
        return boundary.reshape(3, -1)

    def set_arrays(self, arrays):
        """ """
        attributes = self.attributes
        # same length
        dnvert = len(arrays["species_index"]) - len(attributes["species_index"])
        if dnvert > 0:
            self.add_vertices_bmesh(dnvert)
            self.add_vertices_bmesh(dnvert, self.obj_o)
        elif dnvert < 0:
            self.delete_vertices_bmesh(range(-dnvert))
            self.delete_vertices_bmesh(range(-dnvert), self.obj_o)
        if len(self) == 0:
            self.update_mesh()
            return
        self.positions = arrays["positions"][0]
        self.offsets = arrays["offsets"][0]
        self.set_trajectory(arrays)
        species_index = [string2Number(sp) for sp in arrays["species"]]
        self.set_attributes(
            {
                "atoms_index": arrays["atoms_index"],
                "species_index": species_index,
                "scale": arrays["scales"],
                "show": arrays["shows"],
            }
        )
        species = np.unique(arrays["species"])
        for sp in species:
            self.add_geometry_node(sp)

    def get_arrays(self):
        """ """
        object_mode()
        # tstart = time()
        arrays = self.attributes
        arrays.update({"positions": self.positions})
        arrays.update({"offsets": self.offsets})
        # radius
        radius = self.batoms.radius
        arrays.update({"radius": np.zeros(len(self))})
        species = np.array(
            [number2String(i) for i in arrays["species_index"]], dtype="U20"
        )
        arrays["species"] = species
        for sp, value in radius.items():
            mask = np.where(arrays["species"] == sp)
            arrays["radius"][mask] = value
        # size
        arrays["size"] = arrays["radius"] * arrays["scale"]
        # main elements
        main_elements = self.batoms.species.main_elements
        elements = [main_elements[sp] for sp in arrays["species"]]
        arrays.update({"elements": np.array(elements, dtype="U20")})
        # print('get_arrays: %s'%(time() - tstart))
        return arrays

    @property
    def search_bond_data(self):
        return self.get_search_bond_data()

    def get_search_bond_data(self, include_batoms=False):
        """
        using foreach_get and foreach_set to improve performance.
        """
        arrays = self.arrays
        search_bond_data = {
            "positions": arrays["positions"],
            "species": arrays["species"],
            "indices": arrays["atoms_index"],
            "offsets": arrays["offsets"],
        }
        return search_bond_data

    @property
    def offsets(self):
        return self.get_offsets()

    def get_offsets(self):
        """
        using foreach_get and foreach_set to improve performance.
        """
        n = len(self)
        offsets = np.empty(n * 3, dtype=int)
        self.obj_o.data.vertices.foreach_get("co", offsets)
        return offsets.reshape((n, 3))

    @offsets.setter
    def offsets(self, offsets):
        self.set_offsets(offsets)

    def set_offsets(self, offsets):
        """
        Set global offsets to local vertices
        """
        object_mode()
        n = len(self.obj_o.data.vertices)
        if len(offsets) != n:
            raise ValueError("offsets has wrong shape %s != %s." % (len(offsets), n))
        if n == 0:
            return
        offsets = offsets.reshape((n * 3, 1))
        if self.obj_o.data.shape_keys is None and len(self) > 0:
            base_name = "Basis_%s" % self.obj_o.name
            self.obj_o.shape_key_add(name=base_name)
        self.obj_o.data.shape_keys.key_blocks[0].data.foreach_set("co", offsets)
        self.obj_o.data.update()
        self.update_mesh(self.obj_o)

    @property
    def bondlists(self):
        return self.get_bondlists()

    def get_bondlists(self):
        bondlists = self.batoms.bond.arrays
        return bondlists

    def get_trajectory(self):
        """ """
        trajectory = {}
        trajectory["positions"] = self.get_obj_trajectory(self.obj)
        trajectory["offsets"] = self.get_obj_trajectory(self.obj_o)
        return trajectory

    def set_trajectory(self, trajectory=None, frame_start=0):
        if trajectory is None:
            trajectory = self.trajectory
        nframe = len(trajectory)
        if nframe == 0:
            return
        name = "%s_search_bond" % (self.label)
        obj = self.obj
        self.set_shape_key(name, obj, trajectory["positions"], frame_start=frame_start)
        #
        name = "%s_search_bond_offset" % (self.label)
        obj = self.obj_o
        self.set_shape_key(name, obj, trajectory["offsets"], frame_start=frame_start)

    def calc_search_bond_data(self, bondlists, mollists, moldatas, arrays, cell):
        """ """
        # tstart = time()
        # 9th column of bondlists is search bond type 1
        # search type 1: atoms search by bond, the one with offset
        indices1 = np.where((bondlists[:, 2:5] != np.array([0, 0, 0])).any(axis=1))[0]
        indices2 = np.where((bondlists[:, 5:8] != np.array([0, 0, 0])).any(axis=1))[0]
        n = len(indices1) + len(indices2)
        if n == 0:
            return default_search_bond_datas
        bondlists1 = bondlists[indices1]
        bondlists1 = bondlists1[:, [0, 2, 3, 4]]
        bondlists1 = np.unique(bondlists1, axis=0)
        indices1 = bondlists1[:, 0].astype(int)
        model_styles1 = arrays["model_style"][indices1]
        shows1 = arrays["show"][indices1]
        selects1 = arrays["select"][indices1]
        scales1 = arrays["scale"][indices1]
        species_indexs1 = arrays["species_index"][indices1]
        species1 = arrays["species"][indices1]
        offset_vectors1 = bondlists1[:, 1:4]
        offsets1 = np.dot(offset_vectors1, cell)
        positions1 = arrays["positions"][indices1] + offsets1
        # ------------------------------------
        #
        bondlists2 = bondlists[indices2]
        bondlists2 = bondlists2[:, [1, 5, 6, 7]]
        bondlists2 = np.unique(bondlists2, axis=0)
        indices2 = bondlists2[:, 0].astype(int)
        model_styles2 = arrays["model_style"][indices2]
        shows2 = arrays["show"][indices2]
        selects2 = arrays["select"][indices2]
        scales2 = arrays["scale"][indices2]
        species_indexs2 = arrays["species_index"][indices2]
        species2 = arrays["species"][indices2]
        offset_vectors2 = bondlists2[:, 1:4]
        offsets2 = np.dot(offset_vectors2, cell)
        positions2 = arrays["positions"][indices2] + offsets2
        #
        indices = np.append(indices1, indices2)
        species_indexs = np.append(species_indexs1, species_indexs2)
        species = np.append(species1, species2)
        positions = np.append(positions1, positions2, axis=0)
        offset_vectors = np.append(offset_vectors1, offset_vectors2, axis=0)
        model_styles = np.append(model_styles1, model_styles2)
        selects = np.append(selects1, selects2)
        shows = np.append(shows1, shows2)
        scales = np.append(scales1, scales2)
        datas = {
            "atoms_index": np.array(indices),
            "species_index": species_indexs,
            "species": species,
            "positions": np.array([positions]),
            # 'offsets':offsets,
            "offsets": np.array([offset_vectors]),
            "model_styles": model_styles,
            "shows": shows,
            "selects": selects,
            "scales": scales,
        }
        # =========================
        # # search molecule
        # atoms_index = []
        # for b in mollists:
        #     indices = moldatas[b[0]]
        #     datas['atoms_index'] = np.append(datas['atoms_index'], indices)
        #     datas['offsets'] = np.append(datas['offsets'], b[5:8])

        # print('datas: ', datas)
        # print('calc_search_bond_data: {0:10.2f} s'.format(time() - tstart))
        return datas



import bpy
import numpy as np
from batoms.bio.bio import read

try:
    from _common_helpers import has_display, set_cycles_res

    use_cycles = not has_display()
except ImportError:
    use_cycles = False

extras = dict(engine="cycles") if use_cycles else {}


def test_cavity(tio2):
    tio2.boundary = 0.01
    tio2.cavity.resolution = 1
    # tio2.cavity.build_cavity()
    # tio2 *= [2, 2, 2]
    tio2.cavity.draw()
    if use_cycles:
        set_cycles_res(tio2)
    tio2.get_image([0, 1, 0], output="tio2-cavity.png", **extras)


def test_cavity_zsm():
    from batoms.bio.bio import read

    bpy.ops.batoms.delete()
    mof = read("../tests/datas/zsm-5.cif")
    mof.boundary = 0.01
    mof.cavity.resolution = 1
    # mof.cavity.build_cavity()
    mof *= [2, 1, 1]
    mof.cavity.draw()
    if use_cycles:
        set_cycles_res(mof)
    mof.get_image([0, 1, 0], output="mof-5.png", **extras)


def test_cavity_mof():
    from batoms.bio.bio import read

    bpy.ops.batoms.delete()
    mof = read("../tests/datas/mof-5.cif")
    mof.boundary = 0.01
    mof.cavity.resolution = 1
    # mof.cavity.build_cavity()
    mof *= [2, 1, 1]
    mof.cavity.draw()
    if use_cycles:
        set_cycles_res(mof)
    mof.get_image([0, 1, 0], output="mof-5.png", **extras)


def test_cavity_ops():
    bpy.ops.batoms.delete()
    mof = read("../tests/datas/mof-5.cif")
    mof.cavity.resolution = 0.5
    bpy.context.view_layer.objects.active = mof.obj
    bpy.ops.surface.cavity_draw()
    assert len(mof.cavity.settings) == 1
    bpy.ops.surface.cavity_remove(name="0")
    print(mof.cavity.settings)
    assert len(mof.cavity.settings) == 0


def test_gui():
    """latticeplane panel"""
    from batoms.bio.bio import read

    bpy.ops.batoms.delete()
    mof = read("../tests/datas/mof-5.cif")
    bpy.context.view_layer.objects.active = mof.obj
    assert bpy.context.scene.Bcavity.show is True
    bpy.context.scene.Bcavity.show = False
    assert mof.cavity.show is False
    bpy.context.scene.Bcavity.minCave = 3.0
    assert np.isclose(mof.cavity.minCave, 3.0)



"""
"""

import bpy
from bpy.types import Menu, Panel, UIList
from bpy.props import (
    BoolProperty,
    EnumProperty,
)


from batoms.gui.utils import (
    get_active_bpy_data,
    get_attr,
    get_enum_attr,
    set_attr,
    set_enum_attr,
    set_module_attr,
)


model_style_items = [
    ("Surface", "Surface", "", 0),
    ("Dot", "Dot", "", 1),
    ("Wireframe", "Wireframe", "", 2),
]


class CrystalShapeProperties(bpy.types.PropertyGroup):
    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=model_style_items,
        get=get_enum_attr("model_style", get_active_bpy_data("Bcrystalshape")),
        set=set_enum_attr("model_style", set_module_attr("crystal_shape")),
        default=0,
    )

    show: BoolProperty(
        name="show",
        default=False,
        description="show all object for view and rendering",
        get=get_attr("show", get_active_bpy_data("Bcrystalshape")),
        set=set_attr("show", set_module_attr("crystal_shape")),
    )


class VIEW3D_PT_Batoms_crystal_shape(Panel):
    bl_label = "Crystal Shape"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_category = "Plugins"
    bl_idname = "VIEW3D_PT_Batoms_crystal_shape"
    bl_options = {"DEFAULT_CLOSED"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def draw(self, context):
        name = "None"
        if context.object:
            if context.object.batoms.type != "OTHER":
                name = context.object.batoms.label
        layout = self.layout
        layout.label(text="Active: " + name)
        iso = context.scene.Bcrystalshape

        layout.label(text="Model style")
        layout.prop(iso, "model_style", expand=True)
        layout.prop(iso, "show", expand=True)
        layout.separator()


class BATOMS_MT_crystal_shape_context_menu(Menu):
    bl_label = "Crystal Shape Specials"
    bl_idname = "BATOMS_MT_crystal_shape_context_menu"

    def draw(self, _context):
        layout = self.layout
        op = layout.operator(
            "plane.crystal_shape_add", icon="ADD", text="Add Crystal Shape"
        )
        layout.separator()
        op = layout.operator(
            "plane.crystal_shape_remove", icon="X", text="Delete All Crystal Shape"
        )
        op.all = True


class BATOMS_UL_crystal_shape(UIList):
    def draw_item(
        self, _context, layout, _data, item, icon, active_data, _active_propname, index
    ):
        crystal_shape = item
        custom_icon = "OBJECT_DATAMODE"
        if self.layout_type in {"DEFAULT", "COMPACT"}:
            split = layout.split(factor=0.66, align=False)
            split.prop(crystal_shape, "name", text="", emboss=False, icon=custom_icon)
            row = split.row(align=True)
            row.emboss = "NONE_OR_STATUS"
            row.prop(crystal_shape, "distance", text="")
        elif self.layout_type == "GRID":
            layout.alignment = "CENTER"
            layout.label(text="", icon=custom_icon)


class BATOMS_PT_crystal_shape(Panel):
    bl_label = "Crystal Shape"
    bl_category = "Plugins"
    bl_idname = "BATOMS_PT_crystal_shape"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_parent_id = "VIEW3D_PT_Batoms_crystal_shape"
    # bl_options = {'DEFAULT_CLOSED'}

    COMPAT_ENGINES = {"BLENDER_RENDER", "BLENDER_EEVEE", "BLENDER_WORKBENCH"}

    @classmethod
    def poll(cls, context):
        obj = context.object
        if obj:
            return obj.batoms.type != "OTHER"
        else:
            return False

    def draw(self, context):
        layout = self.layout

        ob = context.object
        ba = bpy.data.collections[ob.batoms.label].Bcrystalshape
        if len(ba.settings) > 0:
            kb = ba.settings[ba.ui_list_index]
        else:
            kb = None

        row = layout.row()

        rows = 3
        if kb:
            rows = 5

        row.template_list(
            "BATOMS_UL_crystal_shape",
            "",
            ba,
            "settings",
            ba,
            "ui_list_index",
            rows=rows,
        )

        col = row.column(align=True)
        op = col.operator("plane.crystal_shape_add", icon="ADD", text="")
        op = col.operator("plane.crystal_shape_remove", icon="REMOVE", text="")
        if kb is not None:
            op.name = kb.name
        col.separator()

        col.menu("BATOMS_MT_crystal_shape_context_menu", icon="DOWNARROW_HLT", text="")

        if kb:
            col.separator()

            sub = col.column(align=True)

            split = layout.split(factor=0.4)
            row = split.row()

            row = split.row()
            row.alignment = "RIGHT"

            sub = row.row(align=True)
            sub.label()  # XXX, for alignment only

            sub = row.row()
            layout.use_property_split = True
            row = layout.row()
            col = layout.column()
            sub = col.column(align=True)
            sub.prop(kb, "distance", text="Distance")
            # sub.prop(kb, "scale", text="Scale")
            col.prop(kb, "symmetry", text="Symmetry")
            col.prop(kb, "show_edge", text="Show edge")
            col.prop(kb, "material_style", text="material_style")
            col.prop(kb, "color", text="color")
            op = layout.operator(
                "plane.crystal_shape_draw", icon="GREASEPENCIL", text="Draw"
            )



from shutil import ExecError
import bpy
from mathutils import Vector, Matrix
import numpy as np
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def get_selected_vertices_bmesh(obj):
    """_summary_

    Warnning: bmesh.select_history does not support
     box/lasso/circle/other selected vertices.

    Args:
        obj (bpy.type.object): _description_

    Returns:
        _type_: _description_
    """
    import bmesh

    if obj.batoms.type == "BATOMS" and obj.mode == "EDIT":
        data = obj.data
        bm = bmesh.from_edit_mesh(data)
        v = [s.index for s in bm.select_history if isinstance(s, bmesh.types.BMVert)]
        return v
    else:
        return []


def get_all_consoles():
    """Get all consoles

    Returns:
        list: a list of consoles
    """
    from console_python import get_console

    consoles = []
    for area in bpy.context.screen.areas:
        if area.type == "CONSOLE":
            # areas.append(area)
            for region in area.regions:
                if region.type == "WINDOW":
                    console, stdout, stderr = get_console(hash(region))
                    consoles.append(console)
    return consoles


def object_mode():
    for object in bpy.data.objects:
        if object.mode == "EDIT":
            bpy.ops.object.mode_set(mode="OBJECT")


def eidt_mode():
    try:
        bpy.ops.object.mode_set(mode="EDIT")
    except ExecError:
        logger.critical("Can not change to EDIT mode")


def read_batoms_list():
    """
    Read all batoms collection
    """
    items = [col.name for col in bpy.data.collections if col.batoms.type != "OTHER"]
    return items


def get_selected_batoms():
    """ """
    batoms_list = []
    for obj in bpy.context.selected_objects:
        # for p in ['batom', 'bbond', 'bisosurface', 'bpolyhedra', 'bplane']:
        if obj.batoms.type != "OTHER":
            batoms_list.append(obj.batoms.label)
    batoms_list = list(set(batoms_list))
    # print(batoms_list)
    return batoms_list


def get_selected_objects(attr):
    """ """
    objs = []
    for obj in bpy.context.selected_objects:
        if obj.batoms.type == attr.upper():
            objs.append(obj.name)
    return objs


def get_selected_vertices(obj):
    """ """
    import numpy as np

    selected_vertices = []
    # if obj.mode != "EDIT":
    # logger.warning('Warning: Please switch to Edit mode.')
    # return selected_vertices
    mode = obj.mode
    bpy.ops.object.mode_set(mode="OBJECT")
    count = len(obj.data.vertices)
    sel = np.zeros(count, dtype=bool)
    obj.data.vertices.foreach_get("select", sel)
    selected_vertices = np.where(sel)[0]
    # back to whatever mode we were in
    bpy.ops.object.mode_set(mode=mode)
    return selected_vertices.tolist()


def get_selected_edges(obj):
    """ """
    import numpy as np

    selected_edges = []
    # if obj.mode != "EDIT":
    # logger.warning('Warning: Please switch to Edit mode.')
    # return selected_edges
    mode = obj.mode
    bpy.ops.object.mode_set(mode="OBJECT")
    count = len(obj.data.edges)
    sel = np.zeros(count, dtype=bool)
    obj.data.edges.foreach_get("select", sel)
    selected_edges = np.where(sel)[0]
    # back to whatever mode we were in
    bpy.ops.object.mode_set(mode=mode)
    return selected_edges.tolist()


def get_selected_vertices_all():
    """
    change to 'Object' mode
    in order
    """
    import numpy as np

    eidt_mode()
    selected_vertices = []
    objs = []
    for obj in bpy.context.objects_in_mode:
        if obj.batoms.type != "OTHER":
            objs.append(obj)
    if len(objs) == 0:
        logger.warning("Warning: No atoms is selected, please switch to Edit mode.")
        return selected_vertices
    dict_mode = {}
    for obj in objs:
        dict_mode[obj.name] = obj.mode
    bpy.ops.object.mode_set(mode="OBJECT")
    for obj in objs:
        count = len(obj.data.vertices)
        sel = np.zeros(count, dtype=bool)
        obj.data.vertices.foreach_get("select", sel)
        index = np.where(sel)[0]
        if len(index) > 0:
            selected_vertices.append((obj.batoms.label, index))
    # back to whatever mode we were in
    bpy.ops.object.mode_set(mode="EDIT")
    return selected_vertices


def remove_collection(name, keep_batom=True):
    """

    Note: to avoid crash in macOS, read the name first,
    and then use get method
    """
    collection = bpy.data.collections.get(name)
    objs = collection.all_objects.keys()
    for obj in objs:
        obj = bpy.data.objects.get(obj)
        if keep_batom and obj.batoms.type != "OTHER":
            continue
        bpy.data.objects.remove(obj, do_unlink=True)
    collection = bpy.data.collections.get(name)
    colls = collection.children.keys()
    for coll in colls:
        coll = bpy.data.collections.get(coll)
        bpy.data.collections.remove(coll)
    collection = bpy.data.collections.get(name)
    bpy.data.collections.remove(collection)


def clean_objects_by_name(name):
    obj = bpy.data.objects.get(name)
    bpy.data.objects.remove(obj)


def clean_coll_objects(coll, names=None):
    """
    remove all bond object in the bond collection
    """
    if not names:
        for obj in coll.all_objects:
            bpy.data.objects.remove(obj, do_unlink=True)
    else:
        for name in names:
            for obj in coll.all_objects:
                if name in obj.name:
                    bpy.data.objects.remove(obj, do_unlink=True)


def clean_coll_object_by_type(coll, type):
    """
    remove all objects which have certain type in the given collection.
    """
    objs_name = []
    for obj in coll.all_objects:
        if obj.batoms.type == type:
            objs_name.append(obj.name)
    for obj in objs_name:
        obj = bpy.data.objects.get(obj)
        bpy.data.objects.remove(obj, do_unlink=True)


def removeAll():
    for mesh in bpy.data.meshes:
        bpy.data.meshes.remove(mesh)
    for obj in bpy.data.objects:
        bpy.data.objects.remove(obj)
    for cam in bpy.data.cameras:
        bpy.data.cameras.remove(cam)
    for light in bpy.data.lights:
        bpy.data.lights.remove(light)
    for mat in bpy.data.materials:
        bpy.data.materials.remove(mat)
    for coll in bpy.data.collections:
        if coll.name == "Collection":
            continue
        bpy.data.collections.remove(coll)


def get_keyframes_of_batoms(batoms):
    """
    get keyframes of a batoms
    """
    keyframes = []
    for sp, ba in batoms.items():
        anim = ba.obj.data.animation_data
        if anim is not None and anim.action is not None:
            for fcu in anim.action.fcurves:
                for keyframe in fcu.keyframe_points:
                    x, y = keyframe.co
                    if x not in keyframes:
                        keyframes.append((int(x)))
    return keyframes


def get_shape_key_of_batoms(batoms):
    """
    get shape keys of a batoms
    """
    keyframes = []
    for sp, ba in batoms.items():
        anim = ba.obj.data.animation_data
        if anim is not None and anim.action is not None:
            for fcu in anim.action.fcurves:
                for keyframe in fcu.keyframe_points:
                    x, y = keyframe.co
                    if x not in keyframes:
                        keyframes.append((int(x)))
    return keyframes


def add_keyframe_to_shape_key(obj, attr, values, frames):
    """ """
    for value, frame in zip(values, frames):
        setattr(obj, attr, value)
        obj.keyframe_insert(data_path=attr, frame=frame)


def add_keyframe_visibility(obj, nframe, frame):
    # key as visible on the current frame
    obj.animation_data_create()
    ac = bpy.data.actions.new("Visibility Action")
    obj.animation_data.action = ac
    for data_path in ["hide_viewport", "hide_render"]:
        fc = ac.fcurves.new(data_path=data_path)
        fc.keyframe_points.add(nframe)
        value = [0] * 2 * nframe
        for i in range(nframe):
            value[2 * i] = i
            if i == frame:
                value[2 * i + 1] = 0
            else:
                value[2 * i + 1] = 1
        fc.keyframe_points.foreach_set("co", value)


def set_look_at(obj, target, roll=0):
    """
    Rotate obj to look at target
    """
    if not isinstance(target, Vector):
        target = Vector(target)
    loc = obj.location
    direction = target - loc
    quat = direction.to_track_quat("-Z", "Y")
    quat = quat.to_matrix().to_4x4()
    rollMatrix = Matrix.Rotation(roll, 4, "Z")
    loc = loc.to_tuple()
    obj.matrix_world = quat @ rollMatrix
    obj.location = loc


def lock_to(obj, target=None, location=True, rotation=True):
    """
    track to obj
    """
    if target is not None:
        if location:
            obj.constraints.new(type="COPY_LOCATION")
            obj.constraints["Copy Location"].target = target
        if rotation:
            obj.constraints.new(type="COPY_ROTATION")
            obj.constraints["Copy Rotation"].target = target
    else:
        for c in obj.constraints:
            obj.constraints.remove(c)


def set_world(color=[0.2, 0.2, 0.2, 1.0]):
    """ """
    world = bpy.context.scene.world
    world.use_nodes = True
    node_tree = world.node_tree
    node_tree.nodes["Background"].inputs["Strength"].default_value = 1.0
    node_tree.nodes["Background"].inputs["Color"].default_value = color


def get_node_by_name(nodes, name, type=None):
    node = nodes.get(name)
    if node is None:
        node = nodes.new(type)
        node.name = name
    return node


def set_interface(node_tree, interface):
    """Set interface of a node_tree"""
    for i in interface:
        node_tree.interface.new_socket(name=i[0], socket_type=i[1], in_out=i[2])


def create_node_tree(name, node_group_type="GeometryNodeTree", interface=[]):
    """Create a node_tree by name and type.
    Add default interface"""
    # create a new node_tree if node_tree is None
    node_tree = bpy.data.node_groups.get(name)
    if node_tree is None:
        node_tree = bpy.data.node_groups.new(name, type=node_group_type)
        # add default interface if not exist
        for i in interface:
            if bpy.app.version_string >= "4.0.0":
                node_tree.interface.new_socket(name=i[0], socket_type=i[1], in_out=i[2])
            else:
                if i[2] == "INPUT":
                    node_tree.inputs.new(i[1], i[0])
                elif i[2] == "OUTPUT":
                    node_tree.outputs.new(i[1], i[0])
        # create input and output node
        node_tree.nodes.new("NodeGroupInput")
        node_tree.nodes.new("NodeGroupOutput")
    return node_tree


def get_node_with_node_tree_by_name(
    nodes,
    name,
    node_type="GeometryNodeGroup",
    node_group_type="GeometryNodeTree",
    interface=[],
):
    """Get the node with a node_tree by name.
    If None, Create a new one based on type.
    Return:
        node (bpy.types.Node): A node with a node_tree
    """
    node = nodes.get(name)
    # create a new node if node is None
    if node is None:
        node = nodes.new(type=node_type)
        node.name = name
    node_tree = create_node_tree(name, node_group_type, interface)
    node.node_tree = node_tree
    return node


def clean_default(camera=False, light=True):
    if "Cube" in bpy.data.objects:
        bpy.data.objects.remove(bpy.data.objects["Cube"], do_unlink=True)
    if camera and "Camera" in bpy.data.cameras:
        bpy.data.cameras.remove(bpy.data.cameras["Camera"])
    if light and "Light" in bpy.data.lights:
        bpy.data.lights.remove(bpy.data.lights["Light"])


def show_index():
    """
    show index of vertices in Edit mode
    """
    bpy.context.preferences.view.show_developer_ui = True
    for a in bpy.context.screen.areas:
        if a.type == "VIEW_3D":
            overlay = a.spaces.active.overlay
            overlay.show_extra_indices = True


def get_area(me):
    npoly = len(me.polygons)
    areas = np.zeros(npoly)
    me.polygons.foreach_get("area", areas)
    total_area = np.sum(areas)
    return total_area


def get_volume(me):
    import bmesh

    bm = bmesh.new()
    bm.from_mesh(me)
    volume = bm.calc_volume()
    return volume


def update_object(obj):
    mode = obj.mode
    bpy.context.view_layer.objects.active = obj
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.object.mode_set(mode="OBJECT")
    bpy.ops.object.mode_set(mode=mode)


def hideOneLevel():
    screen = bpy.context.screen
    outliners = [a for a in screen.areas if a.type == "OUTLINER"]
    c = bpy.context.copy()
    for ol in outliners:
        c["area"] = ol
        ol.tag_redraw()
        bpy.ops.outliner.show_one_level(c, open=False)
        ol.tag_redraw()


def build_gn_modifier(obj, name):
    from bl_operators.geometry_nodes import geometry_node_group_empty_new

    modifier = obj.modifiers.new(name=name, type="NODES")
    if bpy.app.version_string >= "4.0.0":
        group = geometry_node_group_empty_new(name)
    else:
        group = geometry_node_group_empty_new()
        group.name = name
    modifier.node_group = group
    return modifier


def get_att_length(mesh, att):
    """get attribute length based on domain

    Args:
        att (bpy.types.Attribute): attribute of a mesh
    Returns:
        int: length of the attribute
    """
    domain = att.domain
    if domain == "POINT":
        n = len(mesh.vertices)
    elif domain == "EDGE":
        n = len(mesh.edges)
    else:
        n = len(mesh.polygons)
    return n


def get_bmesh_domain(bm, att):
    """Get bmesh domain

    Args:
        bm (_type_): _description_
        att (_type_): _description_
    """
    if att.domain == "POINT":
        domain = getattr(bm, "verts")
    elif att.domain == "EDGE":
        domain = getattr(bm, "edges")
    elif att.domain == "FACE":
        domain = getattr(bm, "faces")
    return domain


def get_bmesh_layer(domain, key, dtype):
    """Get bmesh layer

    Args:
        domain (bm.verts, bm.edges, bm.faces): _description_
        key (str): _description_
        dtype (str): _description_
    """
    if dtype == "STRING":
        layer = domain.layers.string.get(key)
    elif dtype == "INT":
        layer = domain.layers.int.get(key)
    elif dtype == "FLOAT":
        layer = domain.layers.float.get(key)
    elif dtype == "FLOAT_VECTOR":
        layer = domain.layers.float_vector.get(key)
    elif dtype == "FLOAT_COLOR":
        layer = domain.layers.float_color.get(key)

    return layer


def set_vertex_color(obj, name, color):
    """Set vertex color

    Args:
        obj (bpy.types.Object): Object to be set
        name (str): name of the color attribute
        color (array): value of color for each vertex
    """
    npoint = len(color)
    mesh = obj.data
    if bpy.app.version_string >= "3.2.0":
        color = color.reshape((npoint * 4, 1))
        mesh.color_attributes.new(name, "FLOAT_COLOR", "POINT")
        mesh.color_attributes[name].data.foreach_set("color", color)
    else:
        import bmesh

        if obj.mode == "EDIT":
            bm = bmesh.from_edit_mesh(obj.data)
            volume_layer = bm.loops.layers.color.new(name)
            for v in bm.verts:
                for loop in v.link_loops:
                    loop[volume_layer] = color[v.index]
            bmesh.update_edit_mesh(obj.data)
        else:
            bm = bmesh.new()
            bm.from_mesh(obj.data)
            volume_layer = bm.loops.layers.color.new(name)
            for v in bm.verts:
                for loop in v.link_loops:
                    loop[volume_layer] = color[v.index]
            bm.to_mesh(obj.data)
            bm.free()


def get_socket_by_identifier(node, identifier, type="inputs"):
    """Get sockets by identifier"""
    for inp in getattr(node, type):
        if inp.identifier == identifier:
            return inp
    return None


# ========================================================
if bpy.app.version_string >= "3.1.0":
    compareNodeType = "FunctionNodeCompare"
else:
    compareNodeType = "FunctionNodeCompareFloats"



import bpy
from bpy.props import PointerProperty

from . import (
    gui_batoms,
    gui_io,
    gui_slicebatoms,
    gui_toolbar,
    gui_cell,
    gui_plane,
    gui_render,
    gui_volumetric_data,
    ui_list_species,
    view3d_mt_batoms_add,
    view3d_mt_object_context_menu,
    view3d_mt_edit_mesh_context_menu,
)


class BatomsCollection(bpy.types.PropertyGroup):
    """
    Collection properties of all panel properties.
    """

    batoms: PointerProperty(type=gui_batoms.BatomsProperties)
    batom: PointerProperty(type=gui_slicebatoms.BatomProperties)
    cell: PointerProperty(type=gui_cell.CellProperties)
    plane: PointerProperty(type=gui_plane.PlaneProperties)
    render: PointerProperty(type=gui_render.RenderProperties)
    io: PointerProperty(type=gui_io.BatomsPropertiesIO)
    volumetric_data: PointerProperty(type=gui_volumetric_data.VolumetricDataProperties)


classes = [
    gui_batoms.Batoms_PT_prepare,
    gui_batoms.BatomsProperties,
    gui_slicebatoms.Batom_PT_prepare,
    gui_slicebatoms.BatomProperties,
    gui_cell.Cell_PT_prepare,
    gui_cell.CellProperties,
    gui_plane.Plane_PT_prepare,
    gui_plane.PlaneProperties,
    gui_plane.AddButton,
    gui_render.Render_PT_prepare,
    gui_render.RenderProperties,
    view3d_mt_batoms_add.VIEW3D_MT_batoms_add,
    view3d_mt_batoms_add.VIEW3D_MT_surface_add,
    view3d_mt_batoms_add.VIEW3D_MT_nanotube_add,
    view3d_mt_batoms_add.VIEW3D_MT_nanoparticle_add,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_batoms_model_style,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_batoms_radius_style,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_batoms_color_style,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_batoms_label,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_batoms,
    view3d_mt_object_context_menu.VIEW3D_MT_object_context_bonds,
    view3d_mt_edit_mesh_context_menu.VIEW3D_MT_edit_mesh_context_batoms_model_style,
    view3d_mt_edit_mesh_context_menu.VIEW3D_MT_edit_mesh_context_batoms,
    ui_list_species.BATOMS_MT_species_context_menu,
    ui_list_species.BATOMS_UL_species,
    ui_list_species.BATOMS_PT_species,
    gui_volumetric_data.VolumetricDataProperties,
    gui_volumetric_data.VIEW3D_PT_Batoms_volumetric_data,
    gui_volumetric_data.BATOMS_MT_volumetric_data_context_menu,
    gui_volumetric_data.BATOMS_UL_volumetric_data,
    gui_volumetric_data.BATOMS_PT_volumetric_data,
    gui_io.BatomsPropertiesIO,
    gui_io.VIEW3D_PT_Batoms_io,
    gui_io.VIEW3D_PT_Batoms_io_materials_project,
    gui_io.Batoms_IO_search_material_project,
    gui_io.VIEW3D_PT_Batoms_io_pubchem,
    gui_io.Batoms_IO_search_pubchem,
    gui_io.VIEW3D_PT_Batoms_io_RSCB,
    gui_io.Batoms_IO_search_RSCB,
    BatomsCollection,
]


def register_class():
    from bpy.utils import register_class

    for cls in classes:
        register_class(cls)
    scene = bpy.types.Scene
    scene.batoms = PointerProperty(type=BatomsCollection)


def unregister_class():
    from bpy.utils import unregister_class

    for cls in reversed(classes):
        unregister_class(cls)
    scene = bpy.types.Scene
    del scene.batoms


def register_menu():
    bpy.types.VIEW3D_MT_add.prepend(view3d_mt_batoms_add.menu_func)
    bpy.types.VIEW3D_MT_object_context_menu.prepend(
        view3d_mt_object_context_menu.menu_func
    )
    bpy.types.VIEW3D_MT_edit_mesh_context_menu.prepend(
        view3d_mt_edit_mesh_context_menu.menu_func
    )


def unregister_menu():
    bpy.types.VIEW3D_MT_add.remove(view3d_mt_batoms_add.menu_func)
    bpy.types.VIEW3D_MT_object_context_menu.remove(
        view3d_mt_object_context_menu.menu_func
    )
    bpy.types.VIEW3D_MT_edit_mesh_context_menu.remove(
        view3d_mt_edit_mesh_context_menu.menu_func
    )


# handle the keymap
wm = bpy.context.window_manager
# Note that in background mode (no GUI available), keyconfigs are not available either,
# so we have to check this to avoid nasty errors in background case.
kc = wm.keyconfigs.addon


def register_keymap():
    # in background mode, we don't need tool and we can not regester keymap
    if kc:
        # tool
        bpy.utils.register_tool(
            gui_toolbar.BatomsTransform,
            after={"builtin.cursor"},
            separator=True,
            group=True,
        )
        bpy.utils.register_tool(
            gui_toolbar.BatomsBoundary, after={gui_toolbar.BatomsTransform.bl_idname}
        )
        bpy.utils.register_tool(
            gui_toolbar.BatomsCell, after={gui_toolbar.BatomsBoundary.bl_idname}
        )
        # bpy.utils.register_tool(gui_toolbar.AddAtoms, after={gui_toolbar.AddSurface.bl_idname})
        bpy.utils.register_tool(
            gui_toolbar.MoleculeEditElement,
            after={"builtin.scale_cage"},
            separator=True,
            group=True,
        )
        bpy.utils.register_tool(
            gui_toolbar.MolecueEditBond,
            after={gui_toolbar.MoleculeEditElement.bl_idname},
            separator=True,
            group=True,
        )


def unregister_keymap():
    if kc:
        bpy.utils.unregister_tool(gui_toolbar.BatomsTransform)
        bpy.utils.unregister_tool(gui_toolbar.BatomsBoundary)
        bpy.utils.unregister_tool(gui_toolbar.BatomsCell)
        bpy.utils.unregister_tool(gui_toolbar.MoleculeEditElement)
        bpy.utils.unregister_tool(gui_toolbar.MolecueEditBond)



"""
Use ASE's build function
https://wiki.fysik.dtu.dk/ase/ase/cluster/cluster.html?highlight=cluster#module-ase.cluster
"""

import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    IntProperty,
    BoolProperty,
)
from ase.cluster import Decahedron, Icosahedron, Octahedron
from batoms import Batoms


class BuildDecahedron(Operator):
    bl_idname = "nano.decahedron_add"
    bl_label = "Add Decahedron"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add decahedron"

    symbol: StringProperty(
        name="Symbol",
        default="Au",
        description="The chemical symbol of the element to use.",
    )

    p: IntProperty(name="p", default=5, min=1, soft_max=10, description="System size")

    q: IntProperty(name="q", default=2, min=1, soft_max=5, description="System size")

    r: IntProperty(name="r", default=0, min=0, soft_max=3, description="System size")

    latticeconstant: IntProperty(
        name="Lattice Constant",
        default=0,
        min=0,
        soft_max=20,
        description="Lattice Constant",
    )

    label: StringProperty(name="Label", default="", description="Label")

    def execute(self, context):
        if self.label == "":
            self.label = self.symbol
        latticeconstant = None if self.latticeconstant == 0 else self.latticeconstant
        atoms = Decahedron(
            symbol=self.symbol,
            p=self.p,
            q=self.q,
            r=self.r,
            latticeconstant=latticeconstant,
        )
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add nanoparticle {}".format(self.label))
        return {"FINISHED"}


class BuildIcosahedron(Operator):
    bl_idname = "nano.icosahedron_add"
    bl_label = "Add Icosahedron"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add Icosahedron"

    symbol: StringProperty(
        name="Symbol",
        default="Au",
        description="The chemical symbol of the element to use.",
    )

    noshells: IntProperty(
        name="noshells", default=5, min=1, soft_max=10, description="System size"
    )

    latticeconstant: IntProperty(
        name="Lattice Constant",
        default=0,
        min=0,
        soft_max=20,
        description="Lattice Constant",
    )

    label: StringProperty(name="Label", default="", description="Label")

    def execute(self, context):
        if self.label == "":
            self.label = self.symbol
        latticeconstant = None if self.latticeconstant == 0 else self.latticeconstant
        atoms = Icosahedron(
            symbol=self.symbol, noshells=self.noshells, latticeconstant=latticeconstant
        )
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add nanoparticle {}".format(self.label))
        return {"FINISHED"}


class BuildOctahedron(Operator):
    bl_idname = "nano.octahedron_add"
    bl_label = "Add Octahedron"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add Octahedron"

    symbol: StringProperty(
        name="Symbol",
        default="Au",
        description="The chemical symbol of the element to use.",
    )

    length: IntProperty(
        name="length",
        default=5,
        min=1,
        soft_max=10,
        description="Number of atoms on the square edges of the complete octahedron.",
    )

    cutoff: IntProperty(
        name="cutoff",
        default=0,
        min=1,
        soft_max=10,
        description="Number of layers cut at each vertex.",
    )

    latticeconstant: IntProperty(
        name="Lattice Constant",
        default=0,
        min=0,
        soft_max=20,
        description="Lattice Constant",
    )

    alloy: BoolProperty(
        name="Alloy", default=False, description="If True the L1_2 structure is used."
    )

    label: StringProperty(name="Label", default="", description="Label")

    def execute(self, context):
        if self.label == "":
            self.label = self.symbol
        latticeconstant = None if self.latticeconstant == 0 else self.latticeconstant
        atoms = Octahedron(
            symbol=self.symbol,
            cutoff=self.cutoff,
            length=self.length,
            latticeconstant=latticeconstant,
            alloy=self.alloy,
        )
        batoms = Batoms(label=self.label, from_ase=atoms)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report({"INFO"}, "Add nanoparticle {}".format(self.label))
        return {"FINISHED"}



import numpy as np
import math
from time import time
import logging

# logger = logging.getLogger('batoms')
logger = logging.getLogger(__name__)


def map_volumetric_data(volumetric_data, coordinates):
    """Interpolate value at given coordinates for
    the volumetric data

    Note: both volumetric data and coordinate are using scaled positions
    inside the cell
    Args:
        volumetric_data (array): volumetric data
        coordinates (array): coordinates

    Returns:
        array: value at given coordinate
    """
    from scipy import ndimage

    # get scaled coordinates
    index = coordinates * volumetric_data.shape
    # map new value
    data = ndimage.map_coordinates(volumetric_data, index.T, order=1)
    return data


def map_color(data, color1=[1, 0, 0, 1], color2=[0, 0, 1, 1]):
    """_summary_

    Args:
        data (_type_): _description_
        color1 (list, optional): _description_. Defaults to [1, 0, 0, 1].
        color2 (list, optional): _description_. Defaults to [0, 0, 1, 1].

    Returns:
        _type_: _description_
    """
    # normalize
    data = (data - np.min(data)) / (np.max(data) - np.min(data))
    # generate color based on value nvc
    color1 = np.array(color1)
    color2 = np.array(color2)
    dcolor = color2 - color1
    color_array = data[:, None] * dcolor
    color = color1 + color_array
    return color


def read_from_others(from_ase=None, from_pymatgen=None, from_pybel=None):
    if from_ase is not None:
        return read_from_ase(from_ase)
    elif from_pymatgen is not None:
        return read_from_pymatgen(from_pymatgen)
    elif from_pybel is not None:
        return read_from_pybel(from_pybel)
    else:
        return None


def read_from_ase(atoms):
    """
    Import structure from ASE atoms.
    """
    if isinstance(atoms, list):
        frames = atoms
        atoms = frames[0]
    else:
        frames = [atoms]
    nframe = len(frames)
    natom = len(atoms)
    arrays = atoms.arrays
    if "species" not in arrays:
        species = np.array(atoms.get_chemical_symbols(), dtype="U20")
    else:
        species = arrays["species"]
    info = atoms.info
    # read frames
    if nframe > 1:
        positions = np.zeros((nframe, natom, 3))
        for i in range(0, nframe):
            positions[i, :, :] = frames[i].positions
    else:
        positions = arrays["positions"]
    # attributes
    attributes = {}
    for key, array in arrays.items():
        if key in ["positions", "species"]:
            continue
        attributes[key] = array = array
    return species, positions, attributes, atoms.cell, npbool2bool(atoms.pbc), info


def read_from_pymatgen(structure):
    """
    Import structure from Pymatgen structure.
    """
    attributes = {}
    symbols = np.array([str(site.specie.symbol) for site in structure])
    natom = len(symbols)
    attributes["species"] = np.array(symbols, dtype="U20")
    if hasattr(structure, "lattice"):
        cell = structure.lattice.matrix
        pbc = True
    else:
        cell = np.zeros(3)
        pbc = False
    positions = [structure[i].coords for i in range(natom)]
    info = {}
    return symbols, positions, attributes, cell, pbc, info


def read_from_pybel(mol):
    """
    conda install openbabel -c conda-forge
    Converts a pybel molecule to batoms.
    Molecules have the following attributes: atoms,
    charge, data, dim, energy, exactmass, formula, molwt,
    spin, sssr, title and unitcell (if crystal data).

    Atoms have the following attributes: atomicmass, atomicnum,
    coords, exactmass, formalcharge, heavyvalence, heterovalence,
    hyb, idx, implicitvalence, isotope, partialcharge, spin, type,
    valence, vector. The .coords attribute provides a tuple (x, y, z)
    of the atom’s coordinates. The remaining attributes are as for
    the Get methods of OBAtom.
    """
    from openbabel import pybel
    from ase.data import chemical_symbols

    number = []
    positions = []
    charges = []
    elements = []
    for atom in mol.atoms:
        number.append(atom.atomicnum)
        elements.append(chemical_symbols[atom.atomicnum])
        positions.append(atom.coords)
        charges.append(atom.formalcharge)
    cell = mol.unitcell if hasattr(mol, "unitcell") else [0, 0, 0]
    pbc = True if hasattr(mol, "unitcell") else False
    species = elements.copy()
    arrays = {"charges": charges}
    bonds = [
        {
            "atoms": [b.GetBeginAtom().GetIndex(), b.GetEndAtom().GetIndex()],
            "order": b.GetBondOrder(),
        }
        for b in pybel.ob.OBMolBondIter(mol.OBMol)
    ]
    return species, positions, arrays, cell, pbc, {"bonds": bonds}


def string2Number(s):
    s = str(s)
    return int.from_bytes(s.encode(), "little")


def number2String(n):
    n = int(n)
    return n.to_bytes(math.ceil(n.bit_length() / 8), "little").decode()


def default_element_prop(element, radius_style="covalent", color_style="JMOL"):
    """
    Get color, radii for element.
    """
    from batoms.data import covalent_radii, vdw_radii
    from ase.data import chemical_symbols
    from batoms.data import jmol_colors, cpk_colors, vesta_color

    element_prop = {}
    number = chemical_symbols.index(element)
    color_style = str(color_style)
    radius_style = str(radius_style)
    color_style = color_style.upper()
    radius_style = radius_style.upper()
    if color_style == "JMOL" or color_style == "0":
        color = jmol_colors[number]
    elif color_style == "CPK" or color_style == "1":
        color = cpk_colors[number]
    elif color_style == "VESTA" or color_style == "2":
        color = vesta_color[element]
    if radius_style == "COVALENT" or radius_style == "0":
        radius = covalent_radii[number]
    elif radius_style == "VDW" or radius_style == "1":
        radius = vdw_radii[number]
    elif radius_style == "IONIC" or radius_style == "2":
        raise Exception("Ionic radii is not supported yet!")
    else:
        raise Exception("radius_style: %s is not supported yet!" % radius_style)
    element_prop["radius"] = radius
    element_prop["color"] = [color[0], color[1], color[2], 1.0]
    return element_prop


def get_default_species_data(
    elements, radius_style="covalent", color_style="JMOL", props={}
):
    """Set default color, radii for elements,

    Args:
        elements (dict): _description_
        radius_style (str, optional): _description_. Defaults to 'covalent'.
        color_style (str, optional): _description_. Defaults to "JMOL".
        props (dict, optional): _description_. Defaults to {}.

    exmaples:
    elements = {"Fe": 0.8, "Cr": 0.2}
    elements = {"Fe": {"occupancy": 0.8}, "Cr": {"occupancy": 0.2}}

    Returns:
        _type_: _description_
    """
    radius = 0
    species_props = {"elements": {}}
    for ele, eledata in elements.items():
        if isinstance(eledata, (int, float)):
            # only has occupancy data
            eledata = {"occupancy": eledata}
        species_props["elements"][ele] = eledata
        data = default_element_prop(
            ele, radius_style=radius_style, color_style=color_style
        )
        radius += data["radius"] * eledata["occupancy"]
        species_props["elements"][ele]["color"] = data["color"]
    species_props["radius"] = radius
    species_props.update(props)
    return species_props


def get_polyhedra_kind(color, width=0.01, show_edge=True, props={}):
    """
    Set initial data for polyhedra.

    """
    polyhedra_kind = {
        "color": color,
        "vertices": [],
        "edges": [],
        "faces": [],
        "show_edge": show_edge,
    }
    polyhedra_kind["edge"] = {
        "lengths": [],
        "centers": [],
        "normals": [],
        "color": (1.0, 1.0, 1.0, 1.0),
        "width": width,
    }
    polyhedra_kind.update(props)
    return polyhedra_kind


def getEquidistantPoints(p1, p2, n):
    points = zip(
        np.linspace(p1[0], p2[0], n + 1),
        np.linspace(p1[1], p2[1], n + 1),
        np.linspace(p1[2], p2[2], n + 1),
    )
    return points


def get_cell_vertices(cell):
    """ """
    cell = np.array(cell)
    cell = cell.reshape(3, 3)
    cell_vertices = np.empty((2, 2, 2, 3))
    for c1 in range(2):
        for c2 in range(2):
            for c3 in range(2):
                cell_vertices[c1, c2, c3] = np.dot([c1, c2, c3], cell)
    cell_vertices.shape = (8, 3)
    return cell_vertices


def get_box(vertices, padding=4):
    """
    Find a minmum box contains all atoms
    """
    # find box
    box = np.zeros((3, 2))
    for i in range(3):
        box[i, 0] = vertices[:, i].min() - padding
        box[i, 1] = vertices[:, i].max() + padding
    return box


def build_grid(box, resolution):
    """
    generate gridpoints
    """
    grids = []
    for i in range(3):
        grids.append(np.arange(box[i, 0], box[i, 1], resolution))
    x, y, z = np.meshgrid(grids[0], grids[1], grids[2], indexing="ij")  # , sparse=True)
    shape = x.shape
    meshgrids = np.c_[x.ravel(), y.ravel(), z.ravel()]
    return meshgrids, shape


def getDistances(points1, points2):
    """Calculate the distance between every point pair in points.
    Parameters
    ----------
    points1: array_like, shape (n1, m)
        m is the dimension of point.
    points2: array_like, shape (n2, m)
        m is the dimension of point.
    Returns
    -------
    indices : array of integers, shape (n1*n2, 2)
        The index of each point in the pairs.
    dist : array of floats, shape (n1*n2)
        The distances of the point pairs.
    """
    # tstart = time()
    npoint1 = len(points1)
    npoint2 = len(points2)
    i, j = np.triu_indices(npoint1, 0, npoint2)
    vectors = points2[j] - points1[i]
    dist = np.linalg.norm(vectors, axis=1)
    indices = np.column_stack([i, j])
    # print("getDistance: {}".format(time() - tstart))
    return indices, dist


def get_canvas(vertices, direction=[0, 0, 1], padding=1):
    """
    Calculate canvas for camera.
    Project vertices and cell on the plane of view.
    Find the boudanry of the object on the plane.
    """

    # plane
    canvas = np.zeros((2, 3))
    frame = rotate_frame(direction)
    #
    projz = np.dot(vertices, frame[2])
    projxy = vertices - projz[:, None] * frame[2]
    projx = np.dot(projxy, frame[0])
    projy = np.dot(projxy, frame[1])
    #
    # find canvas box
    canvas = np.zeros([2, 3])
    canvas[0, 0] = projx.min()
    canvas[1, 0] = projx.max()
    canvas[0, 1] = projy.min()
    canvas[1, 1] = projy.max()
    canvas[0, 2] = projz.min()
    canvas[1, 2] = projz.max()
    canvas[0] -= padding
    canvas[1] += padding
    return canvas


def rotate_frame(direction):
    """
    rotate frame by algin z to direction
    """
    direction = np.array(direction)
    nz = direction / np.linalg.norm(direction)
    nx = np.cross([0, 0, 1], nz) + np.array([1e-6, 0, 0])
    nx = nx / np.linalg.norm(nx)
    ny = np.cross(nz, nx) + np.array([0, 1e-6, 0])
    ny = ny / np.linalg.norm(ny)
    return np.array([nx, ny, nz])


def get_equivalent_indices(no, indices):
    """ """
    from ase.spacegroup import Spacegroup

    sg = Spacegroup(no)
    indices = sg.equivalent_reflections([indices])
    indices = indices.tolist()
    return indices


def get_equivalent_atoms(atoms, tol=1e-5):
    """Get equivalent atoms using spglib."""
    import spglib

    atoms = (atoms.get_cell(), atoms.get_scaled_positions(), atoms.numbers)
    symmetry_data = spglib.get_symmetry_dataset(atoms, symprec=tol)
    return symmetry_data["equivalent_atoms"]


def local2global(positions, matrix, reversed=False):
    """Transferm the positions using (4x4) matrix, so
    one need append one column to the positions.
    The positions could be a Nx3 arrays, or nframxNx3 arrays.
    if reversed is True, from global to local.
    """
    if reversed:
        matrix = np.linalg.inv(matrix)
    if len(positions.shape) == 2:
        positions = np.c_[positions, np.ones(len(positions))]
        positions = np.dot(positions, matrix.T)
        positions = positions[:, :3]
        return positions
    elif len(positions.shape) == 3:
        nframe = len(positions)
        positions = np.c_[positions, np.ones((nframe, len(positions[0]), 1))]
        positions = np.dot(positions, matrix.T)
        positions = positions[:, :, :3]
        return positions


def npbool2bool(pbc):
    """ """
    newpbc = []
    for i in range(3):
        if pbc[i]:
            newpbc.append(True)
        else:
            newpbc.append(False)
    return newpbc


def heron3(a, b, c):
    s = (a + b + c) / 2
    A = np.sqrt(s * (s - a) * (s - b) * (s - c))
    return A


def heron4(u, v, w, U, V, W):
    X = (w - U + v) * (U + v + w)
    x = (U - v + w) * (v - w + U)
    Y = (u - V + w) * (V + w + u)
    y = (V - w + u) * (w - u + V)
    Z = (v - W + u) * (W + u + v)
    z = (W - u + v) * (u - v + W)
    a = np.sqrt(x * Y * Z)
    b = np.sqrt(y * Z * X)
    c = np.sqrt(z * X * Y)
    d = np.sqrt(x * y * z)
    V = np.sqrt((-a + b + c + d) * (a - b + c + d) * (a + b - c + d) * (a + b + c - d))
    V = V / 192 / u / v / w
    return V


def calc_V2(u, v, w, U, V, W):
    a1 = np.square(u)
    a2 = np.square(v)
    a3 = np.square(w)
    a4 = np.square(W)
    a5 = np.square(U)
    a6 = np.square(V)
    V2 = (
        a1 * a5 * (a2 + a3 + a4 + a6 - a1 - a5)
        + a2 * a6 * (a1 + a3 + a4 + a5 - a2 - a6)  # noqa W503
        + a3 * a4 * (a1 + a2 + a5 + a6 - a3 - a4)  # noqa W503
        - a1 * a2 * a4  # noqa W503
        - a2 * a3 * a5  # noqa W503
        - a1 * a3 * a6  # noqa W503
        - a4 * a5 * a6  # noqa W503
    ) / 144
    return V2


def heron42(u, v, w, U, V, W):
    V2 = calc_V2(u, v, w, U, V, W)
    V2 = np.where(V2 > 0, V2, 1e-6)
    Volume = np.sqrt(V2)
    return Volume


def calc_origin_2(p, p0, p1, r0, r1, r):
    xaxis = p1 - p0
    # xaxis = np.sign(xaxis[:, 0])[:, None]*xaxis
    l0 = r0 + r
    l1 = r1 + r
    l2 = np.linalg.norm(xaxis, axis=1)
    #
    vec = p - p0
    xaxis = xaxis / l2[:, None]
    zaxis = np.cross(xaxis, vec)
    zaxis = zaxis / np.linalg.norm(zaxis, axis=1)[:, None]
    yaxis = np.cross(zaxis, xaxis)
    v = (np.square(l0) + np.square(l2) - np.square(l1)) / 2 / l0 / l2
    v[v > 1] = 1
    # print(min(v), max(v))
    angles = np.arccos(v)
    origin = p0 + l0[:, None] * (
        np.cos(angles)[:, None] * xaxis + np.sin(angles)[:, None] * yaxis
    )
    return origin


def check_origin_2(p, p0, p1, r0, r1, r, eps):
    origins = calc_origin_2(p, p0, p1, r0, r1, r)
    vec = p - origins
    norms = np.linalg.norm(vec, axis=1)
    indices = np.where(norms < eps)[0]
    return indices, origins[indices]


def check_origin_3(p, p0, p1, p2, r0, r1, r2, r, eps):
    origins = calc_origin_3(p, p0, p1, p2, r0, r1, r2, r)
    vec = p - origins
    norms = np.linalg.norm(vec, axis=1)
    indices = np.where(norms < eps)[0]
    return indices, origins[indices]


def calc_origin_3(p, p0, p1, p2, r0, r1, r2, r):
    l0 = r0 + r
    l1 = r1 + r
    l2 = r2 + r
    l3 = p1 - p0
    l4 = p2 - p0
    l5 = p2 - p1
    l3 = np.linalg.norm(l3, axis=1)
    l4 = np.linalg.norm(l4, axis=1)
    l5 = np.linalg.norm(l5, axis=1)
    xaxis = p1 - p0
    xaxis = xaxis / np.linalg.norm(xaxis, axis=1)[:, None]
    vec1 = p2 - p0
    zaxis = np.cross(xaxis, vec1)
    zaxis = zaxis / np.linalg.norm(zaxis, axis=1)[:, None]
    yaxis = np.cross(zaxis, xaxis)
    # -------------------------
    V = heron42(l0, l1, l2, l5, l4, l3)
    A = heron3(l3, l4, l5)
    h = 3 * V / A
    angles1 = np.arccos((np.square(l3) + np.square(l0) - np.square(l1)) / 2 / l3 / l0)
    h1 = l0 * np.sin(angles1)
    x = l0 * np.cos(angles1)
    y = np.sqrt(h1 * h1 - h * h)
    # sign of y
    angles2 = np.arccos((np.square(l4) + np.square(l3) - np.square(l5)) / 2 / l4 / l3)
    h2 = np.sqrt(x * x + l4 * l4 - 2 * x * l4 * np.cos(angles2))
    h3 = np.sqrt(l2 * l2 - h * h)
    projy = np.where(h2 > h3, 1, -1) * y
    # sign of z
    vec = p - p0
    projz = np.sign(np.einsum("ij,ij->i", vec, zaxis)) * h
    origin = p0 + x[:, None] * xaxis + projy[:, None] * yaxis + projz[:, None] * zaxis
    return origin


def calc_euler_angle(x, z, seq="xyz"):
    """ """
    from scipy.spatial.transform import Rotation as R

    tstart = time()
    if len(x.shape) == 1:
        x = x.reshape(-1, 3)
    if len(z.shape) == 1:
        z = z.reshape(-1, 3)
    y = np.cross(z, x, axis=1)
    n = len(z)
    if n == 0:
        eulers = np.zeros((0, 3))
    else:
        x = x.reshape(n, 1, 3)
        y = y.reshape(n, 1, 3)
        z = z.reshape(n, 1, 3)
        mat = np.concatenate((x, y, z), axis=1)
        mat = np.linalg.inv(mat)
        r = R.from_matrix(mat)
        eulers = r.as_euler(seq=seq)  # , degrees = True)
    logger.debug("calc_euler_angle: {0:10.2f} s".format(time() - tstart))
    return eulers


def type_blender_to_py(dtype, str="str"):
    """Change Blender data type to python data type

    Args:
        type (_type_): _description_
    """
    type_dict = {
        "INT": "int",
        "INT32_2D": "int",
        "FLOAT": "float",
        "FLOAT2": "float",
        "FLOAT_VECTOR": "float",
        "FLOAT_COLOR": "float",
        "QUATERNION": "float",
        "STRING": str,
        "BOOLEAN": "bool",
    }
    return type_dict[dtype]


def type_py_to_blender(dtype):
    """change python data type to Blender data type

    Args:
        dtype (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Update the subdtype check according to
    # https://github.com/numpy/numpy/blob/db481babcfa7ebc70833e77985858e9295a3135b/numpy/core/numerictypes.py#L357
    if np.issubdtype(dtype, np.integer):
        dtype = "INT"
    elif np.issubdtype(dtype, np.floating):
        dtype = "FLOAT"
    elif np.issubdtype(dtype, np.str_):
        dtype = "STRING"
    elif np.issubdtype(dtype, np.bool_):
        dtype = "BOOLEAN"
    else:
        # raise KeyError('%s is not supported.' % dtype)
        # print('Attribute: {}, {} is not supported.'.format(dtype))
        return False
    return dtype


if __name__ == "__main__":
    # V = heron4(6, 7, 8, 9, 10, 11)
    # V3 = heron4(2.59, 2.92, 2.59, 0.96, 1.52, 0.96)
    # print(V, V2, V3)
    p0 = np.array([[-2.5, 0, 0], [0, 0.7632, -0.4770]])
    p1 = np.array([[0, -0.7632, -0.4770], [0, 0, 0.11926]])
    p2 = np.array([[0, 0, 0.11926], [0, -0.7632, -0.4770]])
    r0 = np.array([1.2, 1.2])
    r1 = np.array([1.2, 1.52])
    r2 = np.array([1.52, 1.2])
    r = 1.4
    p = np.array(
        [[0.89120579, 0.98369867, -0.53714401], [-0.89120579, -0.98369867, 0.53714401]]
    )
    p = np.array(
        [[-0.63896561, 0.47235048, -1.45207691], [-0.80025041, 0.32016024, -1.3009696]]
    )
    # origin = calc_origin(p, p0, p1, r0, r1, r)
    origin = calc_origin_3(p, p0, p1, p2, r0, r1, r2, r)
    print(origin)


def deprecated(msg):
    """Helper function print deprecating warning."""
    logger.warning(msg)
    msg = "=" * 80 + "\n" + "Warning: " + msg + "\n" + "=" * 80 + "\n"
    print(msg)



"""Definition of the plane class.

This module defines the plane object in the Batoms package.

"""

import bpy
import bmesh
import numpy as np
from batoms.plugins.base import PluginObject
from .setting import LatticePlaneSettings
from batoms.draw import draw_cylinder, draw_surface_from_vertices
import logging

logger = logging.getLogger(__name__)


class LatticePlane(PluginObject):
    def __init__(
        self,
        label=None,
        location=np.array([0, 0, 0]),
        batoms=None,
    ):
        """Plane Class

        Args:
            label (_type_, optional): _description_. Defaults to None.
            location (_type_, optional): _description_. Defaults to np.array([0, 0, 0]).
            batoms (_type_, optional): _description_. Defaults to None.
        """
        #
        self.batoms = batoms
        self.label = label
        PluginObject.__init__(self, label, "latticeplane", batoms)
        self.settings = LatticePlaneSettings(self.label, parent=self)
        self.settings.bpy_data.active = True

    def build_materials(
        self,
        name,
        color,
        node_inputs=None,
        material_style="default",
        color_by_attribute=None,
    ):
        """ """
        from batoms.material import create_material

        if name in bpy.data.materials:
            mat = bpy.data.materials.get(name)
            bpy.data.materials.remove(mat, do_unlink=True)
        mat = create_material(
            name,
            color=color,
            node_inputs=node_inputs,
            material_style=material_style,
            backface_culling=False,
            color_by_attribute=color_by_attribute,
        )
        return mat

    def build_plane(self, bcell, include_center=False):
        """
        Build vertices, edges and faces of plane.

        """
        cellEdges = bcell.edges
        planes = {}
        for p in self.settings:
            intersect_points = []
            normal = np.dot(p.indices, bcell.reciprocal)
            normal = normal / np.linalg.norm(normal)
            # get intersection point
            for line in cellEdges:
                point = p.distance * normal
                intersect_point = linePlaneIntersection(line, normal, point)
                if intersect_point is not None:
                    intersect_points.append(intersect_point)
                # get verts, edges, faces by Hull
            if len(intersect_points) < 3:
                continue
            vertices, edges, faces = faces_from_vertices(
                intersect_points, normal, include_center=include_center, scale=p.scale
            )
            planes[p.name] = self.get_plane_data(vertices, edges, faces, p)
        self.planes = planes
        return planes

    def get_plane_data(self, vertices, edges, faces, p):
        """
        build edge
        """
        plane = p.as_dict()
        if len(faces) > 0:
            plane.update(
                {
                    "vertices": vertices,
                    "edges": edges,
                    "faces": faces,
                    "edges_cylinder": {
                        "lengths": [],
                        "centers": [],
                        "normals": [],
                        "vertices": 6,
                        "color": (0.0, 0.0, 0.0, 1.0),
                        "width": p.width,
                        "battr_inputs": {},
                    },
                    "battr_inputs": {"Blatticeplane": p.as_dict()},
                }
            )
            for edge in edges:
                center = (vertices[edge[0]] + vertices[edge[1]]) / 2.0
                vec = vertices[edge[0]] - vertices[edge[1]]
                length = np.linalg.norm(vec)
                nvec = vec / length
                plane["edges_cylinder"]["lengths"].append(length)
                plane["edges_cylinder"]["centers"].append(center)
                plane["edges_cylinder"]["normals"].append(nvec)
        return plane

    def build_slicing(self, name, plane, bcell, cuts=None, cmap="Spectral"):
        """
        Change plane to a 2D slicing plane.
        Use vertex color
        """
        from scipy import ndimage
        from ase.cell import Cell
        from batoms.utils.attribute import set_mesh_attribute

        volume = self.batoms.volumetric_data[plane["color_by"]]
        cell = Cell(bcell.array)
        obj = bpy.data.objects.get(name)
        me = obj.data
        bm = bmesh.new()
        bm.from_mesh(me)
        # get maximum cuts based on the density
        if cuts is None:
            density = bcell.length / volume.shape
            maxlength = 0
            bm.verts.ensure_lookup_table()
            bm.edges.ensure_lookup_table()
            for edge in bm.edges:
                v = bm.verts[edge.verts[0].index].co - bm.verts[edge.verts[1].index].co
                length = np.linalg.norm(v)
                if length > maxlength:
                    maxlength = length
            cuts = maxlength / np.min(density)
            # print('cuts: ', cuts)
        # make mesh by subdivided
        bmesh.ops.subdivide_edges(
            bm,
            edges=bm.edges,
            cuts=int(cuts),
            use_grid_fill=True,
        )
        bm.to_mesh(me)
        me.update()
        # vertices to coordinates
        n = len(me.vertices)
        local_positions = np.empty(n * 3, dtype=np.float64)
        me.vertices.foreach_get("co", local_positions)
        local_positions = local_positions.reshape((n, 3))
        n = len(local_positions)
        # positions (natom, 3) to (natom, 4)
        local_positions = np.append(local_positions, np.ones((n, 1)), axis=1)
        mat = np.array(obj.matrix_world)
        positions = mat.dot(local_positions.T).T
        # (natom, 4) back to (natom, 3)
        positions = positions[:, :3] - bcell.origin
        # get scaled positions
        scaled_positions = cell.scaled_positions(positions)
        index = scaled_positions * volume.shape
        # map new value
        data = ndimage.map_coordinates(volume, index.T, order=1)
        # normalize
        dv = np.max(data) - np.min(data)
        data = (data - np.min(data)) / dv
        # add attribute to mesh
        me.attributes.new(name=plane["color_by"], type="FLOAT", domain="POINT")
        set_mesh_attribute(obj, plane["color_by"], data)
        # bpy.context.view_layer.objects.active = obj
        # bpy.ops.object.mode_set(mode='VERTEX_PAINT')

    def build_slicing_image(self, volume, bcell):
        """
        2D slicings of volumetric data by an arbitrary plane.
        Todo: calculate size
        """
        self.get_symmetry_indices()
        shape = volume.shape
        slicings = {}
        for p in self:
            if not p.slicing:
                continue
            length = bcell.length
            d = p.distance / length
            index = [int(i) for i in d * shape]
            data = []
            if p.indices[0] == 1:
                data = volume[index[0], :, :]
                positions = d * bcell.array[0] + (bcell.array[1] + bcell.array[1]) / 2.0
                rotation = (np.pi / 2, 0, 0)
                size = (length[1], length[2], 1)
            if p.indices[1] == 1:
                data = volume[:, index[1], :]
                positions = d * bcell.array[1] + (bcell.array[0] + bcell.array[2]) / 2.0
                rotation = (0, np.pi / 2, 0)
                size = (length[0], length[2], 1)
            if p.indices[2] == 1:
                data = volume[:, :, index[2]]
                positions = d * bcell.array[2] + (bcell.array[0] + bcell.array[1]) / 2.0
                rotation = (0, 0, 0)
                size = (length[0], length[1], 1)
            imagename = "%s_image_%s.png" % (self.label, p.name)
            save_image(data, imagename, interpolation="bicubic")
            slicings[p.name] = {
                "imagename": imagename,
                "location": positions,
                "rotation": rotation,
                "size": size,
            }
        return slicings

    def build_boundary(self, indices):
        """
        Remove vertices above the plane
        """
        p = self.settings[indices]
        normal = np.dot(np.array(p.indices), self.batoms.cell.reciprocal)
        normal = normal / np.linalg.norm(normal)
        point = p.distance * normal
        # isosurface, plane
        colls = self.batoms.coll.children.keys()
        for coll_name in colls:
            objs = bpy.data.collections.get(coll_name).all_objects.keys()
            if "cell" in coll_name:
                continue
            for obj_name in objs:
                obj = bpy.data.objects.get(obj_name)
                if obj.type != "MESH":
                    continue
                if "volume" in obj.name:
                    continue
                n = len(obj.data.vertices)
                vertices = np.empty(n * 3, dtype=np.float64)
                obj.data.vertices.foreach_get("co", vertices)
                vertices = vertices.reshape((n, 3))
                x1 = np.dot(vertices, normal) - np.dot(point, normal)
                index = np.where(x1 > -1e-6)[0]
                if len(index) == 0:
                    continue
                if obj.batoms.type == "BATOMS":
                    self.batoms.delete(obj.batoms.batom.species, index)
                else:
                    bm = bmesh.new()
                    bm.from_mesh(obj.data)
                    bm.verts.ensure_lookup_table()
                    verts_select = [bm.verts[i] for i in index]
                    bmesh.ops.delete(bm, geom=verts_select, context="VERTS")
                    bm.to_mesh(obj.data)

    def draw(
        self, plane_name="ALL", no=None, cuts=None, cmap="bwr", include_center=False
    ):
        """Draw plane
        no: int
            spacegroup of structure, if None, no will be determined by
            get_spacegroup_number()
        cuts: int
            The number of subdivide for selected plane for 2d slicing
        camp: str,
        default 'bwr'
            colormaps for 2d slicing.
        include_center: bool
            include center of plane in the mesh
        """
        from batoms.utils.butils import clean_coll_object_by_type

        # delete old plane
        clean_coll_object_by_type(self.batoms.coll, "LATTICEPLANE")
        if no is not None:
            self.no = no
        planes = self.build_plane(self.batoms.cell, include_center=include_center)
        for species, plane in planes.items():
            if plane_name.upper() != "ALL" and species != plane_name:
                continue
            if plane["boundary"]:
                name = "%s_%s_%s" % (self.label, self.name, species)
                self.delete_obj(name)
                self.build_boundary(plane["indices"])
                bpy.context.view_layer.update()
            else:
                name = "%s_%s_%s" % (self.label, self.name, species)
                self.delete_obj(name)
                obj = draw_surface_from_vertices(
                    name,
                    plane,
                    coll=self.batoms.coll,
                )
                obj.parent = self.batoms.obj
                obj.batoms.type = "LATTICEPLANE"
                obj.batoms.label = self.batoms.label
                if plane["show_edge"]:
                    name = "%s_%s_%s" % (self.label, "plane_edge", species)
                    self.delete_obj(name)
                    obj = draw_cylinder(
                        name=name,
                        datas=plane["edges_cylinder"],
                        coll=self.batoms.coll,
                    )
                    obj.parent = self.batoms.obj
                    obj.batoms.type = "LATTICEPLANE"
                    obj.batoms.label = self.batoms.label
                if plane["color_by"] != "None":
                    name = "%s_%s_%s" % (self.label, self.name, species)
                    self.build_slicing(
                        name, plane, self.batoms.cell, cuts=cuts, cmap=cmap
                    )
                    color_by_attribute = {
                        "attribute_name": plane["color_by"],
                        "ValToRGB": [plane["color1"], plane["color2"]],
                    }
                else:
                    color_by_attribute = None
                mat = self.build_materials(
                    name,
                    color=plane["color"],
                    material_style=plane["material_style"],
                    color_by_attribute=color_by_attribute,
                )
                obj.data.materials.append(mat)

    @property
    def setting(self):
        from batoms.utils import deprecated

        """setting object."""
        deprecated(
            '"setting" will be deprecated in the furture, please use "settings".'
        )
        return self.settings

    def as_dict(self):
        """ """
        data = {}
        data["settings"] = self.settings.as_dict()
        data.update(self.settings.bpy_data.as_dict())
        return data


def save_image(data, filename, interpolation="bicubic"):
    """ """
    import pylab as plt
    import numpy as np

    data = data.T
    size = np.shape(data)
    # print('size: ', size)
    fig = plt.figure(figsize=(size[1] / size[0] * 10, 10))
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    fig.add_axes(ax)
    ax.imshow(data, interpolation=interpolation)
    plt.savefig(filename, dpi=300)


def faces_from_vertices(vertices, normal, include_center=False, scale=[1, 1, 1]):
    """
    get faces from vertices
    """
    # remove duplicative point
    vertices = np.unique(vertices, axis=0)
    n = len(vertices)
    if n < 3:
        return vertices, [], []
    center = np.mean(vertices, axis=0)
    v1 = vertices[0] - center
    angles = [[0, 0]]
    normal = normal / (np.linalg.norm(normal) + 1e-6)
    for i in range(1, n):
        v2 = vertices[i] - center
        x = np.cross(v1, v2)
        c = np.sign(np.dot(x, normal))
        angle = np.arctan2(c, np.dot(v1, v2))
        angles.append([i, angle])
    # scale
    vec = vertices - center
    # length = np.linalg.norm(vec, axis = 1)
    # nvec = vec/length[:, None]
    vertices = center + np.array([scale]) * vec
    # search convex polyhedra
    angles = sorted(angles, key=lambda x: x[1])
    if not include_center:
        faces = [[a[0] for a in angles]]
        edges = [[angles[0][0], angles[-1][0]]]
        for i in range(0, n - 1):
            edges.append([angles[i][0], angles[i + 1][0]])
    else:
        # add center to vertices
        vertices = np.append(vertices, center.reshape(1, 3), axis=0)
        icenter = len(vertices) - 1
        faces = [[angles[0][0], angles[-1][0], icenter]]
        edges = [
            [angles[0][0], icenter],
            [angles[0][0], angles[-1][0]],
            [angles[-1][0], icenter],
        ]
        for i in range(0, n - 1):
            faces.append([angles[i][0], angles[i + 1][0], icenter])
            edges.extend(
                [
                    [angles[i][0], angles[i + 1][0]],
                    [angles[i][0], icenter],
                    [angles[i + 1][0], icenter],
                ]
            )
    return vertices, edges, faces


def linePlaneIntersection(line, normal, point):
    """
    3D Line Segment and Plane Intersection
    - Point
    - Line contained in plane
    - No intersection
    """
    d = np.dot(point, normal)
    normalLine = line[0] - line[1]
    a = np.dot(normalLine, normal)
    # No intersection or Line contained in plane
    if np.isclose(a, 0):
        return None
    # in same side
    b = np.dot(line, normal) - d
    if b[0] * b[1] > 0:
        return None
    # Point
    v = point - line[0]
    d = np.dot(v, normal) / a
    point = np.round(line[0] + normalLine * d, 6)
    return point


def threePlaneIntersection(planes):
    """
    3D three Planes Intersection
    """
    # Condition for a single point of intersection
    mat = np.concatenate([[plane["normal"]] for plane in planes], axis=0)
    det = np.linalg.det(mat)
    if np.isclose(det, 0):
        return None
    darray = np.array([np.dot(plane["point"], plane["normal"]) for plane in planes])
    point = np.zeros(3)
    for i in range(3):
        temp = mat.copy()
        temp[:, i] = darray
        point[i] = round(np.linalg.det(temp) / det, 6)
    return point


def convexhull(planes, point):
    """
    find points at the same side of origin for all planes
    Todo: '+' or '-'
    """
    if point is None:
        return None
    for name, plane in planes.items():
        x1 = np.dot(point, plane["normal"]) - np.dot(plane["point"], plane["normal"])
        x2 = np.dot(np.array([0, 0, 0]), plane["normal"]) - np.dot(
            plane["point"], plane["normal"]
        )
        if abs(x1) > 1e-6 and x1 * x2 < -1e-6:
            return None
    return point



import bpy
from bpy.types import Operator
from bpy.props import (
    StringProperty,
    BoolProperty,
    FloatVectorProperty,
    FloatProperty,
    IntProperty,
    IntVectorProperty,
    EnumProperty,
)
from batoms import Batoms
from batoms.ops.base import OperatorBatoms, OperatorBatomsEdit
from batoms.utils.butils import get_selected_vertices


class ApplyCell(OperatorBatoms):
    bl_idname = "batoms.apply_cell"
    bl_label = "Apply Cell"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new cell parameters"

    a: FloatVectorProperty(
        name="a",
        default=(1, 0, 0),
        # subtype = "XYZ",
        description="Cell in a axis",
    )
    b: FloatVectorProperty(
        name="b",
        default=(0, 1, 0),
        # subtype = "XYZ",
        description="Cell in b axis",
    )
    c: FloatVectorProperty(
        name="c",
        default=(0, 0, 1),
        # subtype = "XYZ",
        description="Cell in c axis",
    )

    def execute(self, context):
        cell = [self.a, self.b, self.c]
        batoms = Batoms(label=context.object.batoms.label)
        batoms.cell = cell
        batoms.obj.select_set(True)
        return {"FINISHED"}


class ApplyTransform(OperatorBatoms):
    bl_idname = "batoms.apply_transform"
    bl_label = "Apply Transform"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new transform parameters"

    a: IntVectorProperty(
        name="a",
        default=(1, 0, 0, 10),
        size=4,
        soft_min=-10,
        soft_max=10,
        # subtype = "MATRIX",
        description="Transform matrix in a axis",
    )
    b: IntVectorProperty(
        name="b",
        default=(0, 1, 0, 10),
        size=4,
        soft_min=-10,
        soft_max=10,
        # subtype = "XYZ",
        description="Transform matrix in b axis",
    )
    c: IntVectorProperty(
        name="c",
        default=(0, 0, 1, 0),
        size=4,
        soft_min=-10,
        soft_max=10,
        # subtype = "XYZ",
        description="Transform matrix in c axis",
    )

    def execute(self, context):
        transform = [self.a, self.b, self.c]
        batoms = Batoms(label=context.object.batoms.label)
        batoms.transform(transform)
        # batoms.obj.select_set(True)
        # batoms1.obj.select_set(False)
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class ApplyBoundary(OperatorBatoms):
    bl_idname = "batoms.apply_boundary"
    bl_label = "Apply Boundary"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new boundary parameters"

    a: FloatVectorProperty(
        name="a", default=(0, 1), size=2, description="boundary  in a axis"
    )
    b: FloatVectorProperty(
        name="b", default=(0, 1), size=2, description="boundary  in b axis"
    )
    c: FloatVectorProperty(
        name="c", default=(0, 1), size=2, description="boundary  in c axis"
    )

    def execute(self, context):
        boundary = [self.a, self.b, self.c]
        batoms = Batoms(label=context.object.batoms.label)
        batoms.boundary = boundary
        batoms.obj.select_set(True)
        return {"FINISHED"}


class AddSurface(OperatorBatoms):
    bl_idname = "batoms.surface_add"
    bl_label = "Add surface"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add surface"

    label: StringProperty(name="Label", default="surf", description="Label")

    indices: IntVectorProperty(
        name="Miller indices",
        size=3,
        default=(1, 1, 1),
        soft_min=-10,
        soft_max=10,
        description="Miller indices for the plane",
    )

    size: IntVectorProperty(
        name="Size",
        size=3,
        default=(1, 1, 1),
        min=1,
        soft_max=10,
        description="System size in units of the minimal unit cell.",
    )

    vacuum: FloatProperty(
        name="vacuum", default=5.0, min=0, soft_max=15, description="vacuum"
    )

    nlayer: IntProperty(name="layers", min=1, soft_max=10, default=4)

    termination: StringProperty(
        name="termination", default="", description="termination"
    )

    def execute(self, context):
        from ase.build import surface
        from ase.build.surfaces_with_termination import surfaces_with_termination

        batoms = Batoms(label=context.object.batoms.label)
        bulk = batoms.as_ase()
        if len(self.termination) == 0:
            atoms = surface(bulk, self.indices, self.nlayer, self.vacuum)
        else:
            atoms = surfaces_with_termination(
                bulk,
                self.indices,
                self.nlayer,
                self.vacuum,
                termination=self.termination,
            )
        batoms.hide = True
        label = batoms.label + "".join(str(i) for i in self.indices)
        if self.label in bpy.data.collections:
            self.label = "%s.001" % self.label
        batoms = Batoms(label=label, from_ase=atoms)
        batoms = batoms * self.size
        batoms.translate([2, 2, 2])
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class AddRootSurface(OperatorBatoms):
    bl_idname = "batoms.root_surface_add"
    bl_label = "Add root surface"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Add root surface"

    label: StringProperty(name="Label", default="rootsurf", description="Label")

    root: IntProperty(name="root", min=1, soft_max=49, default=3)

    size: IntVectorProperty(
        name="Size",
        size=3,
        default=(1, 1, 1),
        min=1,
        soft_max=10,
        description="System size in units of the minimal unit cell.",
    )

    eps: FloatProperty(name="eps", min=1e-8, soft_max=1e-3, default=1e-6)

    def execute(self, context):
        from ase.build import root_surface

        batoms = Batoms(label=context.object.batoms.label)
        surf = batoms.as_ase()
        atoms = root_surface(surf, self.root, eps=self.eps)
        batoms.hide = True
        label = batoms.label + "_root"
        atoms = atoms * self.size
        if self.label in bpy.data.collections:
            self.label = "%s.001" % self.label
        batoms = Batoms(label=label, from_ase=atoms)
        batoms.translate([2, 2, 2])
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class BatomsJoin(OperatorBatoms):
    bl_idname = "batoms.join"
    bl_label = "Join batoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Join batoms"

    label: StringProperty(name="Label", default="", description="Label")

    def execute(self, context):
        from batoms.utils.butils import get_selected_batoms

        batoms_list = get_selected_batoms()
        if len(batoms_list) < 2:
            self.report({"WARNING"}, "Please select at least two Batoms objects.")
            return {"FINISHED"}
        if self.label == "":
            self.label = batoms_list[0]
        if self.label not in batoms_list:
            self.report({"INFO"}, "Create a new Batoms object: {}.".format(self.label))
        batoms = Batoms(label=self.label)
        for label in batoms_list:
            if self.label == label:
                continue
            batoms += Batoms(label=label)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class BatomsSeparate(OperatorBatoms):
    bl_idname = "batoms.separate"
    bl_label = "Separate batoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Separate batoms based on selects"

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.separate()
        # batoms.obj.select_set(True)
        self.report({"INFO"}, "Separate the Batoms object: {}.".format(self.label))
        return {"FINISHED"}


class deleteBatoms(Operator):
    bl_idname = "batoms.delete"
    bl_label = "Delete batoms"
    bl_description = "Delete batoms"

    label: StringProperty(name="Label", default="", description="Label")

    def execute(self, context):
        from batoms.utils.butils import remove_collection, read_batoms_list

        if self.label != "":
            remove_collection(self.label, keep_batom=False)
        else:
            batoms_list = read_batoms_list()
            for label in batoms_list:
                remove_collection(label, keep_batom=False)
                self.report({"INFO"}, "Delete {}".format(label))
        return {"FINISHED"}


class deleteSelectedBatoms(OperatorBatoms):
    bl_idname = "batoms.delete_selected_batoms"
    bl_label = "Delete selected batoms"
    bl_description = "Delete selected batoms"

    def execute(self, context):
        from batoms.utils.butils import get_selected_batoms, remove_collection

        batoms_list = get_selected_batoms()
        for label in batoms_list:
            remove_collection(label, keep_batom=False)
            self.report({"INFO"}, "Delete {}".format(label))
        return {"FINISHED"}


class ApplyModelStyle(OperatorBatoms):
    bl_idname = "batoms.apply_model_style"
    bl_label = "Apply Model Style"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new model_style parameters"

    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=[
            ("0", "Space-filling", "Use ball and stick"),
            ("1", "Ball-and-stick", "Use ball"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Stick", "Use stick"),
        ],
        default="0",
    )

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.model_style = int(self.model_style)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report(
            {"INFO"},
            "Model style of {} is changed to {}".format(
                context.object.batoms.label, self.model_style
            ),
        )
        return {"FINISHED"}


class ApplyRadiusStyle(OperatorBatoms):
    bl_idname = "batoms.apply_radius_style"
    bl_label = "Apply Radius Style"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new radius_style parameters"

    radius_style: EnumProperty(
        name="radius_style",
        description="Structural models",
        items=[
            ("0", "Covalent", "Covalent"),
            ("1", "VDW", "van der Waals"),
            ("2", "Ionic", "Ionic"),
        ],
        default="0",
    )

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.radius_style = int(self.radius_style)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report(
            {"INFO"},
            "Radius style of {} is changed to {}".format(
                context.object.batoms.label, self.radius_style
            ),
        )
        return {"FINISHED"}


class ApplyColorStyle(OperatorBatoms):
    bl_idname = "batoms.apply_color_style"
    bl_label = "Apply Color Style"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new color_style parameters"

    color_style: EnumProperty(
        name="color_style",
        description="Structural models",
        items=[("0", "JMOL", "JMOL"), ("1", "VESTA", "VESTA"), ("2", "CPK", "CPK")],
        default="0",
    )

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.color_style = int(self.color_style)
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report(
            {"INFO"},
            "Color style of {} is changed to {}".format(
                context.object.batoms.label, self.color_style
            ),
        )
        return {"FINISHED"}


class BatomsCopy(OperatorBatoms):
    bl_idname = "batoms.copy"
    bl_label = "Copy"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Copy a batoms"

    label: StringProperty(
        name="label", default="", description="The label of the copy."
    )

    def execute(self, context):
        if self.label == "":
            return {"FINISHED"}
        batoms = Batoms(label=context.object.batoms.label)
        batoms1 = batoms.copy(self.label)
        # batoms.obj.select_set(True)
        # batoms1.obj.select_set(False)
        bpy.context.view_layer.objects.active = batoms1.obj
        self.report({"INFO"}, "Copy {} to {}".format(batoms.label, self.label))
        return {"FINISHED"}


class ApplyLabel(OperatorBatoms):
    bl_idname = "batoms.apply_label"
    bl_label = "Apply Label"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new label parameters"

    label: StringProperty(name="label", default="", description="Show label")

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.show_label = self.label
        batoms.obj.select_set(True)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report(
            {"INFO"},
            "Show label {} for {}".format(self.label, context.object.batoms.label),
        )
        return {"FINISHED"}


class BatomsAutoBuildSpecies(OperatorBatoms):
    bl_idname = "batoms.auto_build_species"
    bl_label = "Auto build species"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Auto build species"

    tol: FloatProperty(name="tol", default=1e-5, description="tol")

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        batoms.auto_build_species(self.tol)
        bpy.context.view_layer.objects.active = batoms.obj
        self.report(
            {"INFO"}, "Auto build species for {}".format(context.object.batoms.label)
        )
        return {"FINISHED"}


# ===========================================
# Below for Edit mode
# ===========================================


class BatomsReplace(OperatorBatomsEdit):
    bl_idname = "batoms.replace"
    bl_label = "Replace"
    bl_description = "Replace selected atoms by new species"

    species: StringProperty(
        name="species", default="", description="Replaced by this species"
    )

    def execute(self, context):
        if self.species == "":
            return {"FINISHED"}
        obj = context.object
        v = get_selected_vertices(obj)
        batoms = Batoms(label=obj.batoms.label)
        batoms.replace(v, self.species)
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class BatomModify(OperatorBatomsEdit):
    bl_idname = "batoms.batom_modify"
    bl_label = "Modify batom"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Modify batom"

    key: StringProperty(
        name="key", default="scale", description="Replaced by this species"
    )

    scale: FloatProperty(
        name="scale",
        default=0.6,
        min=0,
        soft_max=2,
        description="scale",
    )
    bond: BoolProperty(
        name="Bond",
        default=True,
    )

    show: BoolProperty(
        name="Show",
        default=True,
    )

    def execute(self, context):
        obj = context.object
        v = get_selected_vertices(obj)
        batoms = Batoms(label=obj.batoms.label)
        setattr(batoms[v], self.key, getattr(self, self.key))
        context.view_layer.objects.active = obj
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}


class ApplyModelStyleSelected(OperatorBatoms):
    bl_idname = "batoms.apply_model_style_selected"
    bl_label = "Apply Model Style Selected"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Apply new model_style parameters to selected atoms"

    model_style: EnumProperty(
        name="model_style",
        description="Structural models",
        items=[
            ("0", "Space-filling", "Use ball and stick"),
            ("1", "Ball-and-stick", "Use ball"),
            ("2", "Polyhedral", "Use polyhedral"),
            ("3", "Stick", "Use stick"),
        ],
        default="0",
    )

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        indices = get_selected_vertices(context.object)
        model_style_array = batoms.get_attribute("model_style")
        model_style_array[indices] = int(self.model_style)
        scale_array = batoms.get_attribute("scale")
        if int(self.model_style) == 0:
            scale = 1
        elif int(self.model_style) == 1:
            scale = 0.4
        elif int(self.model_style) == 2:
            scale = 0.4
        elif int(self.model_style) == 3:
            scale = 0.0001
        scale_array[indices] = scale
        batoms.set_model_style_array(model_style_array)
        batoms.set_attributes({"scale": scale_array})
        batoms.obj.select_set(True)
        self.report(
            {"INFO"},
            "Model style of {} atoms are changed to {}".format(
                len(indices), self.model_style
            ),
        )
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class BatomsCopySelected(OperatorBatoms):
    bl_idname = "batoms.copy_selected"
    bl_label = "Copy selected atoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Copy selected atoms"

    select: StringProperty(
        name="select", default="", description="New select for the copied atoms."
    )

    label: StringProperty(
        name="label", default="", description="Label for the new batoms."
    )

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        indices = get_selected_vertices(context.object)
        batoms.obj.select_set(True)
        self.report({"INFO"}, "Copy {} atoms".format(len(indices)))
        bpy.context.view_layer.objects.active = batoms.obj
        return {"FINISHED"}


class BatomsDeleteSelectedAtoms(OperatorBatoms):
    """Delete selected vertices

    Args:
        OperatorBatoms (_type_): _description_

    Returns:
        _type_: _description_
    """

    bl_idname = "batoms.delete_selected_atoms"
    bl_label = "Delete selected atoms"
    bl_options = {"REGISTER", "UNDO"}
    bl_description = "Delete selected atoms"

    def execute(self, context):
        batoms = Batoms(label=context.object.batoms.label)
        indices = get_selected_vertices(context.object)
        del batoms[indices]
        self.report({"INFO"}, "Delete {} atoms".format(len(indices)))
        bpy.context.view_layer.objects.active = batoms.obj
        bpy.ops.object.mode_set(mode="EDIT")
        return {"FINISHED"}



import bpy
from batoms import Batoms


def test_batoms_label_element():
    """Draw label for each atom"""
    import numpy as np

    bpy.ops.batoms.delete()
    bpy.ops.batoms.molecule_add(label="ch4", formula="CH4")
    ch4 = Batoms("ch4")
    ch4.show_label = "index"
    ch4.show_label = None
    ch4.show_label = "species"
    ch4.show_label = "elements"
    ch4.set_attributes({"charges": np.zeros(5)})
    ch4.show_label = "charges"


if __name__ == "__main__":
    test_batoms_label_element()
    print("\n Label: All pass! \n")



bl_info = {
    "name": "Batoms toolbar",
    "author": "Xing Wang",
    "version": (2, 2, 0),
    "blender": (3, 0, 0),
    "location": "File -> Import -> Batoms (xyz, cif, pdb, ...)",
    "description": """Python module for drawing and
rendering atomic structures using blender.""",
    "warning": "",
    "category": "Import-Export",
    "doc_url": "https://beautiful-atoms.readthedocs.io/en/latest/",
    "tracker_url": "https://github.com/beautiful-atoms/beautiful-atoms/issues/new/choose",
}


# install pip dependencies
from .install import pip_dependencies  # noqa: E402

pip_dependencies.install()

from batoms.batoms import Batoms  # noqa: E402


__all__ = ["Batoms"]

from . import (  # noqa: E402
    logger,
    preferences,
    plugins,
    internal_data,
    pip_dependencies,
    ops,
    gui,
    console,
)

logger.set_logger(bl_info["version"])

modules = ["bond", "polyhedra", "render", "ribbon"]


def enable_module():
    import importlib

    for key in modules:
        module = importlib.import_module("batoms.{}".format(key))
        module.register_class()


def disable_module():
    import importlib

    for key in modules:
        module = importlib.import_module("batoms.{}".format(key))
        module.unregister_class()


def register():
    from time import time

    tstart0 = time()
    # dependencies
    pip_dependencies.register_class()
    preferences.register_class()
    # class
    internal_data.register_class()
    # class
    ops.register_class()
    gui.register_class()
    # manual
    ops.register_manual_map()
    # menu
    ops.register_menu()
    gui.register_menu()
    # keymap
    gui.register_keymap()
    # hook
    console.register_hook()
    # modules
    enable_module()
    # plugins
    plugins.enable_plugin()
    logger.root_logger.info("Batoms init time: {:.2f}".format(time() - tstart0))
    logger.update_logging_level()


def unregister():
    # dependencies
    pip_dependencies.unregister_class()
    # class
    internal_data.unregister_class()
    ops.unregister_class()
    gui.unregister_class()
    # manual
    ops.unregister_manual_map()
    # menu
    ops.unregister_menu()
    gui.unregister_menu()
    # keymap
    gui.unregister_keymap()
    # hook
    console.unregister_hook()
    disable_module()
    plugins.disable_plugin()
    preferences.unregister_class()


if __name__ == "__main__":
    register()
